"""Module for processing image files."""
import matplotlib.image as mpimg
from scipy import fftpack, ndimage
import numpy as np
from matplotlib import pyplot as plt

try:
    # from skimage.transform import rotate as sk_rotate
    from skimage.util import random_noise
    from skimage.filters import gaussian

    # from PIL import Image as PIL_Image
except Exception:
    # print("Install skimage, Pillow.", exp)
    pass
# from scipy.ndimage import rotate
import scipy

plt.switch_backend("agg")


class Image(object):
    """Module for processing image files."""

    def __init__(self, values=[], pbc_params=[]):
        """Initialize the module."""
        # pbc_params for 2D image: a,b and gamma
        # 3D : a,b,c, alpha, beta, gamma

        self.values = values  # XY numpy array
        self.pbc_params = pbc_params

    def zoom_interp_2d(self, FFT_image, zoom_factor=40, interpol_factor=1):
        """Zoom and interpolate an image."""
        zoom_size = (FFT_image.shape[0] / zoom_factor) / 2
        window_size_x = FFT_image.shape[0]
        window_size_y = FFT_image.shape[1]

        if np.mod(FFT_image.shape[0] / zoom_factor, 2) == 0:
            F2_zoomed = FFT_image[
                int(window_size_x / 2 - zoom_size) : int(
                    window_size_x / 2 + zoom_size
                ),
                int(window_size_y / 2 - zoom_size) : int(
                    window_size_y / 2 + zoom_size
                ),
            ]
        else:
            F2_zoomed = FFT_image[
                int(window_size_x / 2 - zoom_size) : int(
                    window_size_x / 2 + 1 + zoom_size
                ),
                int(window_size_y / 2 - zoom_size) : int(
                    window_size_y / 2 + 1 + zoom_size
                ),
            ]

        return ndimage.zoom(F2_zoomed, interpol_factor)

    def fourier_transform2D(
        self,
        zoom_factor=30,
        interpol_factor=1,
        use_crop=True,
        pad_bright_spot=True,
    ):
        """Make 2D FT."""
        if use_crop:

            g1 = fftpack.fft2(
                (self.crop_square().rgb_to_gray().values[:, :, 0])
            )
        else:
            g1 = fftpack.fft2((self.rgb_to_gray().values[:, :, 0]))
        g2 = np.abs((fftpack.fftshift((g1))))
        g2 = np.abs((fftpack.fftshift((g1))))
        g3 = self.zoom_interp_2d(
            g2, zoom_factor=zoom_factor, interpol_factor=interpol_factor
        )
        if pad_bright_spot:

            g3 = g3 / np.max(g3)
            x = int(g3.shape[0] / 2)
            y = int(g3.shape[1] / 2)
            g3[x, y] = 0.15

        print("g3 shape", x, y)
        # Zoom in to see the FFT more clearly.

        return Image(values=g3)
        # ft_arr=
        # return ft_arr

    @staticmethod
    def from_file(path):
        """Load image frim file."""
        im = mpimg.imread(path)
        return Image(values=im)
        pass

    def rgb_to_gray(self):
        """Make RGB to Grey scale image."""
        img = np.array(self.values)
        # grayImage = np.zeros(img.shape)
        R = np.array(img[:, :, 0])
        G = np.array(img[:, :, 1])
        B = np.array(img[:, :, 2])

        R = R * 0.299
        G = G * 0.587
        B = B * 0.114

        Avg = R + G + B

        for i in range(3):
            img[:, :, i] = Avg

        return Image(values=img)

    def crop_square(self, size=None):
        """Crop an image."""
        if size is None:
            size = min(self.values.shape[0], self.values.shape[1])
        img_cropped = self.values.copy()[:size, :size, :]

        return Image(values=img_cropped)

    def save(self, filename, interpolation="nearest"):
        """Save an image."""
        # if size is None:
        # from matplotlib import pyplot as plt

        plt.imshow(self.values, interpolation=interpolation)
        plt.savefig(filename)
        plt.close()

    def black_and_white(self, threshold=127):
        """Make black and white image."""
        bw = np.asarray(self.values).copy()
        bw[bw < threshold] = 0
        bw[bw >= threshold] = 255
        return Image(values=bw)

    def gaussian_filter(self, sigma=10):
        """Apply Gaussian filter to an image."""
        sigmax = sigma
        sigmay = sigma

        cy, cx = int(self.values.shape[0] / 2), int(self.values.shape[1] / 2)
        x = np.linspace(0, self.values.shape[0], self.values.shape[0])
        y = np.linspace(0, self.values.shape[1], self.values.shape[1])
        X, Y = np.meshgrid(x, y)
        gmask = np.exp(-(((X - cx) / sigmax) ** 2 + ((Y - cy) / sigmay) ** 2))
        new_values = self.values * gmask
        return Image(values=new_values)

    def rotate(self, angle=45):
        """Rotate an image."""
        rot = scipy.ndimage.rotate(self.values, angle)
        return Image(values=rot)

    @staticmethod
    def crop_from_center(
        image_path="stm_image.png",
        image_arr=None,
        target_left=512,
        target_right=512,
        greyscale=True,
    ):
        """Crop squarre from an image."""
        # For STM image, use min_size=50

        from PIL import Image as PIL_Image

        if image_arr is not None:
            if greyscale:
                pil_image = PIL_Image.fromarray(
                    image_arr.astype("uint8")
                ).convert("L")
            else:
                pil_image = PIL_Image.fromarray(image_arr.astype("uint8"))
        else:
            if greyscale:
                pil_image = PIL_Image.open(image_path).convert("L")
            else:
                pil_image = PIL_Image.open(image_path)
        left = int(pil_image.size[0] / 2 - target_left / 2)
        upper = int(pil_image.size[1] / 2 - target_right / 2)
        right = left + target_left
        lower = upper + target_right

        im_cropped = pil_image.crop((left, upper, right, lower))
        return np.asarray(im_cropped)

    @staticmethod
    def augment_image(
        image_path="stm_image.png",
        image_arr=None,
        wrap="wrap",
        rotation_angles=[30, 45, 60],
        rand_noise_var=[0.05, 0.1, 0.5],
        sigma=[5],
        target_left=512,
        target_right=512,
        multichannel=True,
        use_crop=True,
        save_figures=True,
        greyscale=True,
        prefix="example",
        suffix=".png",
    ):
        """Augment images using skimage."""
        from skimage.transform import rotate as sk_rotate

        if image_arr is None:
            img_arr = plt.imread(image_path)
        else:
            img_arr = image_arr

        images = [img_arr]
        for i in rotation_angles:
            rot = sk_rotate(img_arr, angle=i, mode=wrap)
            images.append(rot)
        lr = np.fliplr(img_arr)
        images.append(lr)
        ud = np.flipud(img_arr)
        images.append(ud)
        for i in rand_noise_var:
            rand = random_noise(img_arr, var=i)
            images.append(rand)
        for i in sigma:
            blurred = gaussian(img_arr, sigma=i, multichannel=True)
            images.append(blurred)
        if use_crop:
            images = [
                Image.crop_from_center(
                    image_arr=255 * i,
                    greyscale=greyscale,
                    target_left=target_left,
                    target_right=target_right,
                )
                for i in images
            ]
        if save_figures:
            for ii, i in enumerate(images):
                name = prefix + "_" + str(ii) + suffix
                if greyscale:
                    plt.imshow(i / 255, cmap="gray")
                    plt.tight_layout()
                    plt.axis("off")
                    plt.savefig(name)
                else:
                    plt.imshow(i)
                    plt.tight_layout()
                    plt.axis("off")
                    plt.savefig(name)
        return images

    @staticmethod
    def get_blob_angles(
        filename="image.png",
        tol=0.005,
        tmp_path="tmp.png",
        param1=100,
        param2=255,
        s1=3,
        s2=20,
        rad=10,
        color="r",
        out_name="final.png",
    ):
        """Get angles between atoms."""
        try:
            import cv2
        except ImportError as error:
            print(error.__class__.__name__ + ": " + error.message)
        # https://www.geeksforgeeks.org/
        # white-and-black-dot-detection-using-opencv-python/
        plt.clf()
        im = Image.from_file(filename)
        vals = im.black_and_white(tol).values
        plt.imshow(vals, cmap="jet", interpolation="none")
        plt.axis("off")
        plt.savefig(tmp_path)
        plt.close()
        path = tmp_path
        # reading the image in grayscale mode
        gray = cv2.imread(path, 0)
        # reading the image in grayscale mode
        th, threshed = cv2.threshold(
            gray, param1, param2, cv2.THRESH_BINARY | cv2.THRESH_OTSU
        )
        # findcontours
        cnts = cv2.findContours(
            threshed, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
        )[-2]

        # filter by area
        xcnts = []

        for cnt in cnts:
            if s1 < cv2.contourArea(cnt) < s2:
                xcnts.append(cnt)
        canvas = gray
        cv2.drawContours(canvas, xcnts, -1, (0, 255, 0), 1)
        fig, ax = plt.subplots()
        plt.imshow(canvas)
        plt.savefig(out_name)
        plt.close()
        my_coords = []
        for i in xcnts:

            circle1 = plt.Circle((i[0][0][0], i[0][0][1]), rad, color=color)
            my_coords.append([i[0][0][0], i[0][0][1]])
            ax.add_patch(circle1)
        #

        # printing output
        print("\nDots number: {}".format(len(xcnts)))

        def angle_between(p1, p2):
            ang1 = np.arctan2(*p1[::-1])
            ang2 = np.arctan2(*p2[::-1])
            return np.rad2deg((ang1 - ang2) % (2 * np.pi))

        all_angle_data = []
        for i in my_coords:
            for j in my_coords:
                if i != j:
                    all_angle_data.append([i, j, angle_between(i, j)])
                    print(i, j, angle_between(i, j))
        return all_angle_data


"""
if __name__ == "__main__":
    from jarvis.db.figshare import make_stm_from_prev_parchg

    make_stm_from_prev_parchg()
    # Image.get_blob_angles(filename='stm_image.png')
    # p = "JVASP-667_neg.jpg"
    # im = Image.from_file(p)
    im = Image.crop_from_center()
    ims = Image.augment_image()
    print(ims)
    # plt.imshow(
    #    im.fourier_transform2D(use_crop=True, zoom_factor=50)
    #    .rotate(angle=0)
    #    .black_and_white(threshold=0.05)
    #    .values,
    #    cmap="Greys",
    # )
"""



"""Module to run Tc calculation."""
from jarvis.io.qe.outputs import DataFileSchema
from jarvis.core.atoms import Atoms
from jarvis.core.kpoints import Kpoints3D
from jarvis.tasks.qe.qe import QEjob

# import numpy as np
import os

# from jarvis.core.utils import get_factors


def very_clean():
    """Clean files."""
    cmd = (
        "rm -r elph_dir _ph0 *.dos* *.in *.json  *.fc *.freq* *.save "
        + "*.out *.dyn* *wfc* *.xml  *save lambda *.modes dyn*"
    )
    os.system(cmd)


class SuperCond(object):
    """Module to calculate Tc."""

    def __init__(self, atoms=None, kp=None, qp=None, qe_cmd="pw.x"):
        """Initialize the class."""
        self.atoms = atoms
        self.kp = kp
        self.qp = qp
        self.qe_cmd = qe_cmd

    def to_dict(self):
        """Get dictionary."""
        info = {}
        info["atoms"] = self.atoms.to_dict()
        info["kp"] = self.kp.to_dict()
        info["qp"] = self.qp.to_dict()
        info["qe_cmd"] = self.qe_cmd
        return info

    @classmethod
    def from_dict(self, info={}):
        """Load from a dictionary."""
        return SuperCond(
            atoms=Atoms.from_dict(info["atoms"]),
            kp=Kpoints3D.from_dict(info["kp"]),
            qp=Kpoints3D.from_dict(info["qp"]),
            qe_cmd=info["qe_cmd"],
        )

    def runjob(self):
        """Calculate Tc using QE."""
        # Still under development,
        atoms = self.atoms
        kp = self.kp
        qp = self.qp
        relax = {
            "control": {
                # "calculation": "'scf'",
                "calculation": "'vc-relax'",
                "restart_mode": "'from_scratch'",
                "prefix": "'RELAX'",
                "outdir": "'./'",
                "tstress": ".true.",
                "tprnfor": ".true.",
                "disk_io": "'low'",
                "wf_collect": ".true.",
                "pseudo_dir": None,
                "verbosity": "'high'",
                "nstep": 100,
            },
            "system": {
                "ibrav": 0,
                "nat": None,
                "ntyp": None,
                "ecutwfc": 45,
                "ecutrho": 250,
                "q2sigma": 1,
                "ecfixed": 44.5,
                "qcutz": 800,
                "occupations": "'smearing'",
                "degauss": 0.01,
                "lda_plus_u": ".false.",
            },
            "electrons": {
                "diagonalization": "'david'",
                "mixing_mode": "'local-TF'",
                "mixing_beta": 0.3,
                "conv_thr": "1d-10",
            },
            "ions": {"ion_dynamics": "'bfgs'"},
            "cell": {"cell_dynamics": "'bfgs'", "cell_dofree": "'all'"},
        }
        qejob_relax = QEjob(
            atoms=atoms,
            input_params=relax,
            output_file="relax.out",
            qe_cmd=self.qe_cmd,
            jobname="relax",
            kpoints=kp,
            input_file="arelax.in",
        )

        info = qejob_relax.runjob()
        xml_path = info["xml_path"]
        xml_dat = DataFileSchema(xml_path)
        final_strt = xml_dat.final_structure

        print("final_strt", final_strt)

        # scf_init =
        scf_init = {
            "control": {
                "calculation": "'scf'",
                "restart_mode": "'from_scratch'",
                "prefix": "'QE'",
                "outdir": "'./'",
                "tstress": ".true.",
                "tprnfor": ".true.",
                "disk_io": "'low'",
                "pseudo_dir": None,
                "verbosity": "'high'",
                "nstep": 100,
                "etot_conv_thr": "1.0d-5",
                "forc_conv_thr": "1.0d-4",
            },
            "system": {
                "ibrav": 0,
                # "degauss": 0.01,
                "nat": None,
                "ntyp": None,
                "ecutwfc": 45,
                "ecutrho": 250,
                "occupations": "'tetrahedra_opt'",
                # "smearing": "'mp'",
                # "la2F ": ".true.",
            },
            "electrons": {
                "diagonalization": "'david'",
                "mixing_mode": "'plain'",
                "mixing_beta": 0.7,
                "conv_thr": "1d-10",
            },
        }

        qejob_scf_init = QEjob(
            atoms=final_strt,
            input_params=scf_init,
            output_file="scf_init.out",
            qe_cmd=self.qe_cmd,
            jobname="scf_init",
            kpoints=kp,
            input_file="ascf_init.in",
        )

        info_scf = qejob_scf_init.runjob()
        print(info_scf)
        kpts = kp._kpoints[0]
        qpts = qp._kpoints[0]
        nq1 = qpts[0]  # get_factors(kpts[0])[0]
        nq2 = qpts[1]  # get_factors(kpts[1])[0]
        nq3 = qpts[2]  # get_factors(kpts[2])[0]
        nk1 = kpts[0]
        nk2 = kpts[1]
        nk3 = kpts[2]
        ph = {
            "inputph": {
                "prefix": "'QE'",
                "fildyn": "'QE.dyn'",
                "outdir": "'./'",
                "ldisp": ".true.",
                "lshift_q": ".true.",
                "fildvscf": "'dvscf'",
                "fildrho": "'dvrho'",
                # "electron_phonon": "'lambda_tetra'",
                # "electron_phonon": "'interpolated'",
                # "el_ph_sigma": 0.005,
                "nq1": nq1,
                "nq2": nq2,
                "nq3": nq3,
                "nk1": nk1,
                "nk2": nk2,
                "nk3": nk3,
                # "tr2_ph": "1.0d-12",
            },
        }
        qejob_ph = QEjob(
            atoms=final_strt,
            input_params=ph,
            output_file="ph.out",
            qe_cmd=self.qe_cmd.replace("pw.x", "ph.x"),
            jobname="ph",
            kpoints=None,
            input_file="aph.in",
        )

        qejob_ph.runjob()

        nq1 = qpts[0]  # get_factors(kpts[0])[0]
        nq2 = qpts[1]  # get_factors(kpts[1])[0]
        nq3 = qpts[2]  # get_factors(kpts[2])[0]
        nk1 = kpts[0]
        nk2 = kpts[1]
        nk3 = kpts[2]
        ph_tetra = {
            "inputph": {
                "prefix": "'QE'",
                "fildyn": "'QE.dyn'",
                "outdir": "'./'",
                "ldisp": ".true.",
                "lshift_q": ".true.",
                "fildvscf": "'dvscf'",
                "fildrho": "'dvrho'",
                "electron_phonon": "'lambda_tetra'",
                # "electron_phonon": "'interpolated'",
                # "el_ph_sigma": 0.005,
                "nq1": nq1,
                "nq2": nq2,
                "nq3": nq3,
                "nk1": nk1,
                "nk2": nk2,
                "nk3": nk3,
                # "tr2_ph": "1.0d-12",
            },
            "inputa2f": {
                "nfreq": 500,
            },
        }
        qejob_ph_tetra = QEjob(
            atoms=final_strt,
            input_params=ph_tetra,
            output_file="ph_tetra.out",
            qe_cmd=self.qe_cmd.replace("pw.x", "ph.x"),
            jobname="ph_tetra",
            kpoints=None,
            input_file="aph_tetra.in",
        )

        qejob_ph_tetra.runjob()
        cmd = (
            self.qe_cmd.replace("pw.x", "alpha2f.x")
            + "<"
            + "aph_tetra.in"
            + ">"
            + "aph_tetra.out"
        )
        os.system(cmd)



from jarvis.io.vasp.outputs import Vasprun, Outcar
from jarvis.analysis.phonon.ir import ir_intensity, ir_intensity_phonopy
import os
import shutil

out = Outcar(
    os.path.join(
        os.path.dirname(__file__), "..", "..", "io", "vasp", "OUTCAR.JVASP-39"
    )
)
vrun_file = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "io",
    "vasp",
    "vasprun.xml.JVASP-39",
)
dirc = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "io",
    "vasp",
)

vrun = Vasprun(vrun_file)


def test_ir():
    phonon_eigenvectors = vrun.dfpt_data["phonon_eigenvectors"]
    vrun_eigs = vrun.dfpt_data["phonon_eigenvalues"]
    phonon_eigenvalues = out.phonon_eigenvalues
    masses = vrun.dfpt_data["masses"]
    born_charges = vrun.dfpt_data["born_charges"]
    x, y = ir_intensity(
        phonon_eigenvectors=phonon_eigenvectors,
        phonon_eigenvalues=phonon_eigenvalues,
        masses=masses,
        born_charges=born_charges,
    )
    print(max(y))
    assert round(max(y), 2) == round(0.3511482090386446, 2)
    pdos = vrun.partial_dos_spdf
    cwd = os.getcwd()
    if not os.path.exists("vasprun.xml"):
        shutil.copy2(vrun_file, "vasprun.xml")
    x, y = ir_intensity_phonopy(vasprun=vrun_file)  # , run_dir=dirc)


test_ir()



"""Modules for making point-defect vacancies."""
import pprint
from collections import OrderedDict
from jarvis.analysis.structure.spacegroup import Spacegroup3D
# from jarvis.core.utils import rand_select
from jarvis.core.atoms import Atoms

# import numpy as np
import random


class Vacancy(object):
    """Obtain vacancy defects in Atoms class using Wyckoff data."""

    def __init__(
        self,
        atoms=None,
        defect_structure=None,
        defect_index=None,
        wyckoff_multiplicity=None,
        symbol=None,
    ):
        """
        Initialize the method.

        Arguments are given below.
        Args:
            atoms: jarvis.core.Atoms object.
            defect_index: atoms index for defect.
            defect_structure:  Atoms with defect.
            wyckoff_multiplicity: Wyckoff multiplicity.
            symbol: Elemenyt symbol.
        """
        self._atoms = atoms
        self._defect_index = defect_index
        self._defect_structure = defect_structure
        self._wyckoff_multiplicity = wyckoff_multiplicity
        self._symbol = symbol

    @classmethod
    def from_dict(self, d={}):
        """Load from a dictionary."""
        return Vacancy(
            atoms=Atoms.from_dict(d["atoms"]),
            defect_structure=Atoms.from_dict(d["defect_structure"]),
            defect_index=d["defect_index"],
            wyckoff_multiplicity=d["wyckoff_multiplicity"],
            symbol=d["symbol"],
        )

    def generate_defects(
        self,
        enforce_c_size=10.0,
        on_conventional_cell=False,
        extend=1,
        using_wyckoffs=True,
    ):
        """Provide function to generate defects."""
        atoms = self._atoms
        if on_conventional_cell:
            atoms = Spacegroup3D(atoms).conventional_standard_structure
        a = atoms.lattice.lat_lengths()[0]
        b = atoms.lattice.lat_lengths()[1]
        c = atoms.lattice.lat_lengths()[2]
        if enforce_c_size is not None:
            dim1 = int(float(enforce_c_size) / float(a)) + extend
            dim2 = int(float(enforce_c_size) / float(b)) + extend
            dim3 = int(float(enforce_c_size) / float(c)) + extend
            # atoms = atoms.make_supercell([dim1, dim2, dim3])
            if dim1 == 0:
                dim1 = 1
            if dim2 == 0:
                dim2 = 1
            if dim3 == 0:
                dim3 = 1
            supercell_size = [dim1, dim2, dim3]
        if using_wyckoffs:
            spg = Spacegroup3D(atoms)
            wyckoffs = spg._dataset["wyckoffs"]
            atoms.props = wyckoffs
        else:
            wyckoffs = ["a" for i in range(atoms.num_atoms)]
            atoms.props = wyckoffs
        supercell = atoms.make_supercell(supercell_size)
        # props = rand_select(supercell.props)
        new_props = {}
        for ii, jj, kk in zip(
            supercell.props, supercell.elements, range(supercell.num_atoms)
        ):
            if ii + "_" + jj not in new_props:
                new_props[ii + "_" + jj] = kk
        # print ('props',props)
        # print ('newprops',new_props)
        vacs = []
        for i, j in new_props.items():
            defect_strt = supercell.remove_site_by_index(j)
            vac = Vacancy(
                atoms=supercell,
                defect_structure=defect_strt,
                defect_index=j,
                wyckoff_multiplicity=i,
                symbol=supercell.elements[j],
            )
            vacs.append(vac)
        return vacs

    def to_dict(self):
        """Convert to a dictionary."""
        d = OrderedDict()
        d["atoms"] = self._atoms.to_dict()
        if self._defect_structure is not None:
            d["defect_structure"] = self._defect_structure.to_dict()
        else:
            d["defect_structure"] = None
        d["defect_index"] = self._defect_index
        d["wyckoff_multiplicity"] = self._wyckoff_multiplicity
        d["symbol"] = self._symbol
        return d

    def __repr__(self, indent=2):
        """Representation of the class as dict."""
        return pprint.pformat(self.to_dict(), indent=indent)


def generate_random_defects(n_vacs=10, atoms=None, element=None, seed=123):
    """Generate random defects for an element."""
    # natoms = atoms.num_atoms
    atoms = atoms.to_dict()
    elements = atoms["elements"]
    if element is None:
        element = elements[0]
    coords = atoms["coords"]
    lattice_mat = atoms["lattice_mat"]
    # ids = np.arange(natoms)
    new_elements = []
    new_coords = []
    options = []
    for ii, i in enumerate(elements):
        if i == element:
            options.append(ii)
    random.seed(seed)
    random.shuffle(options)
    to_delete = options[0:n_vacs]
    for ii, i in enumerate(elements):
        if ii not in to_delete:
            new_elements.append(i)
            new_coords.append(coords[ii])
    new_atoms = Atoms(
        coords=new_coords,
        lattice_mat=lattice_mat,
        elements=new_elements,
        cartesian=False,
    )
    return new_atoms



"""Module to run Tc calculation."""
# Ref:
# https://www.nature.com/articles/s41524-022-00933-1
from jarvis.io.qe.outputs import DataFileSchema
from jarvis.core.atoms import Atoms
from jarvis.core.kpoints import Kpoints3D
from jarvis.tasks.qe.qe import QEjob
from jarvis.tasks.qe.converg import converg_kpoints
import numpy as np
import os
from jarvis.core.utils import get_factors

# from jarvis.core.utils import get_factors


def calc_Tc(wlog=300, lamb=1.0, mu=0.1):
    """Calculate Tc."""
    tc = (wlog / 1.2) * np.exp(
        -1.04 * (1 + lamb) / (lamb * (1 - 0.62 * mu) - mu)
    )
    return tc


def parse_lambda(filename="lambda"):
    """Parse lambda file."""
    f = open(filename, "r")
    lines = f.read().splitlines()
    f.close()
    for i in lines:
        if "Broadening" in i:
            tmp = i.split()
            print(i.split())
            wlog = float(tmp[-1])
            lamb = float(tmp[3])
            Tc = calc_Tc(wlog=wlog, lamb=lamb)
            print("Tc", Tc)
            print()


def non_prime_kpoints(kpts=[]):
    """Get non prime kpoints."""
    mem = []
    for i in kpts:
        facts = get_factors(i)
        if len(facts) == 1:
            val = i + 1
        else:
            val = i
        mem.append(val)
    return mem


def very_clean():
    """Clean files."""
    cmd = (
        "rm -r elph_dir _ph0 *.dos* *.in *.json  *.fc *.freq* *.save "
        + "*.out *.dyn* *wfc* *.xml  *save lambda *.modes dyn*"
    )
    os.system(cmd)


class SuperCond(object):
    """Module to calculate Tc."""

    def __init__(
        self,
        atoms=None,
        kp=None,
        qp=None,
        qe_cmd="pw.x",
        relax_calc="'vc-relax'",
        pressure=None,
        psp_dir=None,
        url=None,
        psp_temp_name=None,
        clean_files=True,
    ):
        """Initialize the class."""
        self.atoms = atoms
        self.kp = kp
        self.qp = qp
        # self.kp = kp
        # self.qp = qp
        self.relax_calc = relax_calc
        self.qe_cmd = qe_cmd
        self.psp_dir = psp_dir
        self.url = url
        self.pressure = pressure
        self.psp_temp_name = psp_temp_name
        self.clean_files = clean_files

    def to_dict(self):
        """Get dictionary."""
        info = {}
        info["atoms"] = self.atoms.to_dict()

        info["kp"] = self.kp.to_dict()
        info["qp"] = self.qp.to_dict()
        info["qe_cmd"] = self.qe_cmd
        info["psp_dir"] = self.psp_dir
        info["psp_temp_name"] = self.psp_temp_name
        info["url"] = self.url
        info["relax_calc"] = self.relax_calc
        info["pressure"] = self.pressure
        return info

    @classmethod
    def from_dict(self, info={}):
        """Load from a dictionary."""
        return SuperCond(
            atoms=Atoms.from_dict(info["atoms"]),
            kp=Kpoints3D.from_dict(info["kp"]),
            qp=Kpoints3D.from_dict(info["qp"]),
            qe_cmd=info["qe_cmd"],
            pressure=info["pressure"],
            relax_calc=info["relax_calc"],
            psp_dir=info["psp_dir"],
            url=info["url"],
            psp_temp_name=info["psp_temp_name"],
        )

    def runjob(self):
        """Calculate Tc using QE."""
        # Still under development,
        atoms = self.atoms
        kp = self.kp
        qp = self.qp
        if not kp._kpoints:
            kp_len = converg_kpoints(
                atoms=atoms,
                qe_cmd=self.qe_cmd,
                psp_dir=self.psp_dir,
                url=self.url,
                psp_temp_name=self.psp_temp_name,
            )
            kp = Kpoints3D().automatic_length_mesh(
                lattice_mat=atoms.lattice_mat, length=kp_len
            )
            kpts = kp._kpoints[0]
            kpts = non_prime_kpoints(kpts)
            kp = Kpoints3D(kpoints=[kpts])
            print("kpts", kpts)
        else:
            kpts = kp._kpoints[0]
            kpts = non_prime_kpoints(kpts)
        print("kpts xyz", kpts)
        nq1 = get_factors(kpts[0])[0]
        nq2 = get_factors(kpts[1])[0]
        nq3 = get_factors(kpts[2])[0]
        qp = Kpoints3D(kpoints=[[nq1, nq2, nq3]])
        self.kp = kp
        self.qp = qp
        relax = {
            "control": {
                # "calculation": "'scf'",
                "calculation": self.relax_calc,  # "'vc-relax'",
                "restart_mode": "'from_scratch'",
                "prefix": "'RELAX'",
                "outdir": "'./'",
                "tstress": ".true.",
                "tprnfor": ".true.",
                "disk_io": "'nowf'",
                "wf_collect": ".true.",
                "pseudo_dir": None,
                "verbosity": "'high'",
                "nstep": 100,
            },
            "system": {
                "ibrav": 0,
                "nat": None,
                "ntyp": None,
                "ecutwfc": 45,
                "ecutrho": 250,
                "q2sigma": 1,
                "ecfixed": 44.5,
                "qcutz": 800,
                "occupations": "'smearing'",
                "degauss": 0.01,
                "lda_plus_u": ".false.",
            },
            "electrons": {
                "diagonalization": "'david'",
                "mixing_mode": "'local-TF'",
                "mixing_beta": 0.3,
                "conv_thr": "1d-9",
            },
            "ions": {"ion_dynamics": "'bfgs'"},
            "cell": {"cell_dynamics": "'bfgs'", "cell_dofree": "'all'"},
        }
        if self.pressure is not None:
            relax["cell"]["press"] = self.pressure
        qejob_relax = QEjob(
            atoms=atoms,
            input_params=relax,
            output_file="relax.out",
            qe_cmd=self.qe_cmd,
            jobname="relax",
            kpoints=kp,
            input_file="arelax.in",
            url=self.url,
            psp_dir=self.psp_dir,
            psp_temp_name=self.psp_temp_name,
        )

        info = qejob_relax.runjob()
        xml_path = info["xml_path"]
        xml_dat = DataFileSchema(xml_path)
        final_strt = xml_dat.final_structure

        print("final_strt", final_strt)

        # scf_init =
        scf_init = {
            "control": {
                "calculation": "'scf'",
                "restart_mode": "'from_scratch'",
                "prefix": "'QE'",
                "outdir": "'./'",
                "tstress": ".true.",
                "tprnfor": ".true.",
                "disk_io": "'low'",
                # "disk_io": "'nowf'",
                "pseudo_dir": None,
                "verbosity": "'high'",
                "nstep": 100,
                "etot_conv_thr": "1.0d-5",
                "forc_conv_thr": "1.0d-4",
            },
            "system": {
                "ibrav": 0,
                "degauss": 0.01,
                "nat": None,
                "ntyp": None,
                "ecutwfc": 45,
                "ecutrho": 250,
                "occupations": "'smearing'",
                "smearing": "'mp'",
                "la2F ": ".true.",
            },
            "electrons": {
                "diagonalization": "'david'",
                "mixing_mode": "'plain'",
                "mixing_beta": 0.7,
                "conv_thr": "1d-9",
            },
        }

        qejob_scf_init = QEjob(
            atoms=final_strt,
            input_params=scf_init,
            output_file="scf_init.out",
            qe_cmd=self.qe_cmd,
            jobname="scf_init",
            kpoints=kp,
            input_file="ascf_init.in",
            url=self.url,
            psp_temp_name=self.psp_temp_name,
            psp_dir=self.psp_dir,
        )

        info_scf = qejob_scf_init.runjob()
        print(info_scf)
        # kpts = kp._kpoints[0]
        qpts = qp._kpoints[0]
        nq1 = qpts[0]  # get_factors(kpts[0])[0]
        nq2 = qpts[1]  # get_factors(kpts[1])[0]
        nq3 = qpts[2]  # get_factors(kpts[2])[0]
        ph = {
            "inputph": {
                "prefix": "'QE'",
                "fildyn": "'QE.dyn'",
                "outdir": "'./'",
                "ldisp": ".true.",
                "trans": ".true.",
                "fildvscf": "'dvscf'",
                "electron_phonon": "'interpolated'",
                "el_ph_sigma": 0.005,
                "nq1": nq1,
                "nq2": nq2,
                "nq3": nq3,
                "tr2_ph": "1.0d-12",
            }
        }
        qejob_ph = QEjob(
            atoms=final_strt,
            input_params=ph,
            output_file="ph.out",
            qe_cmd=self.qe_cmd.replace("pw.x", "ph.x"),
            jobname="ph",
            kpoints=None,
            input_file="aph.in",
            url=self.url,
            psp_temp_name=self.psp_temp_name,
            psp_dir=self.psp_dir,
        )

        qejob_ph.runjob()
        # import sys
        # sys.exit()
        qr = {
            "input": {
                "zasr": "'simple'",
                "fildyn": "'QE.dyn'",
                "flfrc": "'QE333.fc'",
                "la2F": ".true.",
            }
        }
        qejob_qr = QEjob(
            atoms=final_strt,
            input_params=qr,
            output_file="q2r.out",
            # qe_cmd="/home/knc6/Software/qe/q-e/bin/q2r.x",
            qe_cmd=self.qe_cmd.replace("pw.x", "q2r.x"),
            jobname="qr",
            kpoints=None,
            input_file="aqr.in",
            url=self.url,
            psp_temp_name=self.psp_temp_name,
            psp_dir=self.psp_dir,
        )

        qejob_qr.runjob()

        kpts = kp._kpoints[0]
        matdyn = {
            "input": {
                "asr": "'simple'",
                "flfrc": "'QE333.fc'",
                "flfrq": "'QE333.freq'",
                "la2F": ".true.",
                "dos": ".true.",
                "fldos": "'phonon.dos'",
                "nk1": kpts[0],
                "nk2": kpts[1],
                "nk3": kpts[2],
                "ndos": 50,
            }
        }

        qejob_matdyn = QEjob(
            atoms=final_strt,
            input_params=matdyn,
            output_file="matdyn.out",
            # qe_cmd="/home/knc6/Software/qe/q-e/bin/matdyn.x",
            qe_cmd=self.qe_cmd.replace("pw.x", "matdyn.x"),
            jobname="matdyn",
            kpoints=None,
            input_file="amatdyn.in",
            url=self.url,
            psp_temp_name=self.psp_temp_name,
            psp_dir=self.psp_dir,
        )

        qejob_matdyn.runjob()
        parse_lambda()
        cmd = "rm -r _ph0 *.wfc* */wfc* */*UPF */*upf */charge-density.dat"
        if self.clean_files:
            os.system(cmd)



"""Module to generate networkx graphs."""

from jarvis.core.atoms import get_supercell_dims
from jarvis.core.specie import Specie
from jarvis.core.utils import random_colors
import numpy as np
import pandas as pd
from collections import OrderedDict
from jarvis.analysis.structure.neighbors import NeighborsAnalysis
from jarvis.core.specie import chem_data, get_node_attributes

# from jarvis.core.atoms import Atoms
from collections import defaultdict
from typing import List, Tuple, Sequence, Optional
import torch
from tqdm import tqdm
import dgl


def canonize_edge(
    src_id,
    dst_id,
    src_image,
    dst_image,
):
    """Compute canonical edge representation.

    Sort vertex ids
    shift periodic images so the first vertex is in (0,0,0) image
    """
    # store directed edges src_id <= dst_id
    if dst_id < src_id:
        src_id, dst_id = dst_id, src_id
        src_image, dst_image = dst_image, src_image

    # shift periodic images so that src is in (0,0,0) image
    if not np.array_equal(src_image, (0, 0, 0)):
        shift = src_image
        src_image = tuple(np.subtract(src_image, shift))
        dst_image = tuple(np.subtract(dst_image, shift))

    assert src_image == (0, 0, 0)

    return src_id, dst_id, src_image, dst_image


def nearest_neighbor_edges(
    atoms=None,
    cutoff=8,
    max_neighbors=12,
    id=None,
    use_canonize=True,
    # use_canonize=False,
):
    """Construct k-NN edge list."""
    # returns List[List[Tuple[site, distance, index, image]]]
    all_neighbors = atoms.get_all_neighbors(r=cutoff)

    # if a site has too few neighbors, increase the cutoff radius
    min_nbrs = min(len(neighborlist) for neighborlist in all_neighbors)

    attempt = 0
    # print ('cutoff=',all_neighbors)
    if min_nbrs < max_neighbors:
        # print("extending cutoff radius!", attempt, cutoff, id)
        lat = atoms.lattice
        if cutoff < max(lat.a, lat.b, lat.c):
            r_cut = max(lat.a, lat.b, lat.c)
        else:
            r_cut = 2 * cutoff
        attempt += 1

        return nearest_neighbor_edges(
            atoms=atoms,
            use_canonize=use_canonize,
            cutoff=r_cut,
            max_neighbors=max_neighbors,
            id=id,
        )
    # build up edge list
    # NOTE: currently there's no guarantee that this creates undirected graphs
    # An undirected solution would build the full edge list where nodes are
    # keyed by (index, image), and ensure each edge has a complementary edge

    # indeed, JVASP-59628 is an example of a calculation where this produces
    # a graph where one site has no incident edges!

    # build an edge dictionary u -> v
    # so later we can run through the dictionary
    # and remove all pairs of edges
    # so what's left is the odd ones out
    edges = defaultdict(set)
    for site_idx, neighborlist in enumerate(all_neighbors):

        # sort on distance
        neighborlist = sorted(neighborlist, key=lambda x: x[2])
        distances = np.array([nbr[2] for nbr in neighborlist])
        ids = np.array([nbr[1] for nbr in neighborlist])
        images = np.array([nbr[3] for nbr in neighborlist])

        # find the distance to the k-th nearest neighbor
        max_dist = distances[max_neighbors - 1]
        # max_dist = distances[max_neighbors - 1]

        # keep all edges out to the neighbor shell of the k-th neighbor
        ids = ids[distances <= max_dist]
        images = images[distances <= max_dist]
        distances = distances[distances <= max_dist]

        # keep track of cell-resolved edges
        # to enforce undirected graph construction
        for dst, image in zip(ids, images):
            src_id, dst_id, src_image, dst_image = canonize_edge(
                site_idx, dst, (0, 0, 0), tuple(image)
            )
            if use_canonize:
                edges[(src_id, dst_id)].add(dst_image)
            else:
                edges[(site_idx, dst)].add(tuple(image))

    return edges


def build_undirected_edgedata(
    atoms=None,
    edges={},
):
    """Build undirected graph data from edge set.

    edges: dictionary mapping (src_id, dst_id) to set of dst_image
    r: cartesian displacement vector from src -> dst
    """
    # second pass: construct *undirected* graph
    # import pprint
    u, v, r = [], [], []
    for (src_id, dst_id), images in edges.items():

        for dst_image in images:
            # fractional coordinate for periodic image of dst
            dst_coord = atoms.frac_coords[dst_id] + dst_image
            # cartesian displacement vector pointing from src -> dst
            d = atoms.lattice.cart_coords(
                dst_coord - atoms.frac_coords[src_id]
            )
            # if np.linalg.norm(d)!=0:
            # print ('jv',dst_image,d)
            # add edges for both directions
            for uu, vv, dd in [(src_id, dst_id, d), (dst_id, src_id, -d)]:
                u.append(uu)
                v.append(vv)
                r.append(dd)

    u, v, r = (np.array(x) for x in (u, v, r))
    u = torch.tensor(u)
    v = torch.tensor(v)
    r = torch.tensor(r).type(torch.get_default_dtype())

    return u, v, r


class Graph(object):
    """Generate a graph object."""

    def __init__(
        self,
        nodes=[],
        node_attributes=[],
        edges=[],
        edge_attributes=[],
        color_map=None,
        labels=None,
    ):
        """
        Initialize the graph object.

        Args:
            nodes: IDs of the graph nodes as integer array.

            node_attributes: node features as multi-dimensional array.

            edges: connectivity as a (u,v) pair where u is
                   the source index and v the destination ID.

            edge_attributes: attributes for each connectivity.
                             as simple as euclidean distances.
        """
        self.nodes = nodes
        self.node_attributes = node_attributes
        self.edges = edges
        self.edge_attributes = edge_attributes
        self.color_map = color_map
        self.labels = labels

    @staticmethod
    def atom_dgl_multigraph(
        atoms=None,
        neighbor_strategy="k-nearest",
        cutoff=8.0,
        max_neighbors=12,
        atom_features="cgcnn",
        max_attempts=3,
        id: Optional[str] = None,
        compute_line_graph: bool = True,
        use_canonize: bool = True,
        # use_canonize: bool = False,
    ):
        """Obtain a DGLGraph for Atoms object."""
        if neighbor_strategy == "k-nearest":
            edges = nearest_neighbor_edges(
                atoms=atoms,
                cutoff=cutoff,
                max_neighbors=max_neighbors,
                id=id,
                use_canonize=use_canonize,
            )
        else:
            raise ValueError("Not implemented yet", neighbor_strategy)
        # elif neighbor_strategy == "voronoi":
        #    edges = voronoi_edges(structure)

        u, v, r = build_undirected_edgedata(atoms, edges)

        # build up atom attribute tensor
        sps_features = []
        for ii, s in enumerate(atoms.elements):
            feat = list(get_node_attributes(s, atom_features=atom_features))
            # if include_prdf_angles:
            #    feat=feat+list(prdf[ii])+list(adf[ii])
            sps_features.append(feat)
        sps_features = np.array(sps_features)
        node_features = torch.tensor(sps_features).type(
            torch.get_default_dtype()
        )
        g = dgl.graph((u, v))
        g.ndata["atom_features"] = node_features
        g.edata["r"] = r

        if compute_line_graph:
            # construct atomistic line graph
            # (nodes are bonds, edges are bond pairs)
            # and add bond angle cosines as edge features
            lg = g.line_graph(shared=True)
            lg.apply_edges(compute_bond_cosines)
            return g, lg
        else:
            return g

    @staticmethod
    def from_atoms(
        atoms=None,
        get_prim=False,
        zero_diag=False,
        node_atomwise_angle_dist=False,
        node_atomwise_rdf=False,
        features="basic",
        enforce_c_size=10.0,
        max_n=100,
        max_cut=5.0,
        verbose=False,
        make_colormap=True,
    ):
        """
        Get Networkx graph. Requires Networkx installation.

        Args:
             atoms: jarvis.core.Atoms object.

             rcut: cut-off after which distance will be set to zero
                   in the adjacency matrix.

             features: Node features.
                       'atomic_number': graph with atomic numbers only.
                       'cfid': 438 chemical descriptors from CFID.
                       'cgcnn': hot encoded 92 features.
                       'basic':10 features
                       'atomic_fraction': graph with atomic fractions
                                         in 103 elements.
                       array: array with CFID chemical descriptor names.
                       See: jarvis/core/specie.py

             enforce_c_size: minimum size of the simulation cell in Angst.
        """
        if get_prim:
            atoms = atoms.get_primitive_atoms
        dim = get_supercell_dims(atoms=atoms, enforce_c_size=enforce_c_size)
        atoms = atoms.make_supercell(dim)

        adj = np.array(atoms.raw_distance_matrix.copy())

        # zero out edges with bond length greater than threshold
        adj[adj >= max_cut] = 0

        if zero_diag:
            np.fill_diagonal(adj, 0.0)
        nodes = np.arange(atoms.num_atoms)
        if features == "atomic_number":
            node_attributes = np.array(
                [[np.array(Specie(i).Z)] for i in atoms.elements],
                dtype="float",
            )
        if features == "atomic_fraction":
            node_attributes = []
            fracs = atoms.composition.atomic_fraction_array
            for i in fracs:
                node_attributes.append(np.array([float(i)]))
            node_attributes = np.array(node_attributes)

        elif features == "basic":
            feats = [
                "Z",
                "coulmn",
                "row",
                "X",
                "atom_rad",
                "nsvalence",
                "npvalence",
                "ndvalence",
                "nfvalence",
                "first_ion_en",
                "elec_aff",
            ]
            node_attributes = []
            for i in atoms.elements:
                tmp = []
                for j in feats:
                    tmp.append(Specie(i).element_property(j))
                node_attributes.append(tmp)
            node_attributes = np.array(node_attributes, dtype="float")
        elif features == "cfid":
            node_attributes = np.array(
                [np.array(Specie(i).get_descrp_arr) for i in atoms.elements],
                dtype="float",
            )
        elif isinstance(features, list):
            node_attributes = []
            for i in atoms.elements:
                tmp = []
                for j in features:
                    tmp.append(Specie(i).element_property(j))
                node_attributes.append(tmp)
            node_attributes = np.array(node_attributes, dtype="float")
        else:
            print("Please check the input options.")
        if node_atomwise_rdf or node_atomwise_angle_dist:
            nbr = NeighborsAnalysis(
                atoms, max_n=max_n, verbose=verbose, max_cut=max_cut
            )
        if node_atomwise_rdf:
            node_attributes = np.concatenate(
                (node_attributes, nbr.atomwise_radial_dist()), axis=1
            )
            node_attributes = np.array(node_attributes, dtype="float")
        if node_atomwise_angle_dist:
            node_attributes = np.concatenate(
                (node_attributes, nbr.atomwise_angle_dist()), axis=1
            )
            node_attributes = np.array(node_attributes, dtype="float")

        # construct edge list
        uv = []
        edge_features = []
        for ii, i in enumerate(atoms.elements):
            for jj, j in enumerate(atoms.elements):
                bondlength = adj[ii, jj]
                if bondlength > 0:
                    uv.append((ii, jj))
                    edge_features.append(bondlength)

        edge_attributes = edge_features

        if make_colormap:
            sps = atoms.uniq_species
            color_dict = random_colors(number_of_colors=len(sps))
            new_colors = {}
            for i, j in color_dict.items():
                new_colors[sps[i]] = j
            color_map = []
            for ii, i in enumerate(atoms.elements):
                color_map.append(new_colors[i])
        return Graph(
            nodes=nodes,
            edges=uv,
            node_attributes=np.array(node_attributes),
            edge_attributes=np.array(edge_attributes),
            color_map=color_map,
        )

    def to_networkx(self):
        """Get networkx representation."""
        import networkx as nx

        graph = nx.DiGraph()
        graph.add_nodes_from(self.nodes)
        graph.add_edges_from(self.edges)
        for i, j in zip(self.edges, self.edge_attributes):
            graph.add_edge(i[0], i[1], weight=j)
        return graph

    @property
    def num_nodes(self):
        """Return number of nodes in the graph."""
        return len(self.nodes)

    @property
    def num_edges(self):
        """Return number of edges in the graph."""
        return len(self.edges)

    @classmethod
    def from_dict(self, d={}):
        """Constuct class from a dictionary."""
        return Graph(
            nodes=d["nodes"],
            edges=d["edges"],
            node_attributes=d["node_attributes"],
            edge_attributes=d["edge_attributes"],
            color_map=d["color_map"],
            labels=d["labels"],
        )

    def to_dict(self):
        """Provide dictionary representation of the Graph object."""
        info = OrderedDict()
        info["nodes"] = np.array(self.nodes).tolist()
        info["edges"] = np.array(self.edges).tolist()
        info["node_attributes"] = np.array(self.node_attributes).tolist()
        info["edge_attributes"] = np.array(self.edge_attributes).tolist()
        info["color_map"] = np.array(self.color_map).tolist()
        info["labels"] = np.array(self.labels).tolist()
        return info

    def __repr__(self):
        """Provide representation during print statements."""
        return "Graph({})".format(self.to_dict())

    @property
    def adjacency_matrix(self):
        """Provide adjacency_matrix of graph."""
        A = np.zeros((self.num_nodes, self.num_nodes))
        for edge, a in zip(self.edges, self.edge_attributes):
            A[edge] = a
        return A


class Standardize(torch.nn.Module):
    """Standardize atom_features: subtract mean and divide by std."""

    def __init__(self, mean: torch.Tensor, std: torch.Tensor):
        """Register featurewise mean and standard deviation."""
        super().__init__()
        self.mean = mean
        self.std = std

    def forward(self, g: dgl.DGLGraph):
        """Apply standardization to atom_features."""
        g = g.local_var()
        h = g.ndata.pop("atom_features")
        g.ndata["atom_features"] = (h - self.mean) / self.std
        return g


def prepare_dgl_batch(
    batch: Tuple[dgl.DGLGraph, torch.Tensor], device=None, non_blocking=False
):
    """Send batched dgl crystal graph to device."""
    g, t = batch
    batch = (
        g.to(device, non_blocking=non_blocking),
        t.to(device, non_blocking=non_blocking),
    )

    return batch


def prepare_line_graph_batch(
    batch: Tuple[Tuple[dgl.DGLGraph, dgl.DGLGraph], torch.Tensor],
    device=None,
    non_blocking=False,
):
    """Send line graph batch to device.

    Note: the batch is a nested tuple, with the graph and line graph together
    """
    g, lg, t = batch
    batch = (
        (
            g.to(device, non_blocking=non_blocking),
            lg.to(device, non_blocking=non_blocking),
        ),
        t.to(device, non_blocking=non_blocking),
    )

    return batch


# def prepare_batch(batch, device=None):
#     """Send tuple to device, including DGLGraphs."""
#     return tuple(x.to(device) for x in batch)


def compute_bond_cosines(edges):
    """Compute bond angle cosines from bond displacement vectors."""
    # line graph edge: (a, b), (b, c)
    # `a -> b -> c`
    # use law of cosines to compute angles cosines
    # negate src bond so displacements are like `a <- b -> c`
    # cos(theta) = ba \dot bc / (||ba|| ||bc||)
    r1 = -edges.src["r"]
    r2 = edges.dst["r"]
    bond_cosine = torch.sum(r1 * r2, dim=1) / (
        torch.norm(r1, dim=1) * torch.norm(r2, dim=1)
    )
    bond_cosine = torch.clamp(bond_cosine, -1, 1)
    # bond_cosine = torch.arccos((torch.clamp(bond_cosine, -1, 1)))
    # print (r1,r1.shape)
    # print (r2,r2.shape)
    # print (bond_cosine,bond_cosine.shape)
    return {"h": bond_cosine}


class StructureDataset(torch.utils.data.Dataset):
    """Dataset of crystal DGLGraphs."""

    def __init__(
        self,
        df: pd.DataFrame,
        graphs: Sequence[dgl.DGLGraph],
        target: str,
        atom_features="atomic_number",
        transform=None,
        line_graph=False,
        classification=False,
        id_tag="jid",
    ):
        """Pytorch Dataset for atomistic graphs.

        `df`: pandas dataframe from e.g. jarvis.db.figshare.data
        `graphs`: DGLGraph representations corresponding to rows in `df`
        `target`: key for label column in `df`
        """
        self.df = df
        self.graphs = graphs
        self.target = target
        self.line_graph = line_graph

        self.labels = self.df[target]
        self.ids = self.df[id_tag]
        self.labels = torch.tensor(self.df[target]).type(
            torch.get_default_dtype()
        )
        self.transform = transform

        features = self._get_attribute_lookup(atom_features)

        # load selected node representation
        # assume graphs contain atomic number in g.ndata["atom_features"]
        for g in graphs:
            z = g.ndata.pop("atom_features")
            g.ndata["atomic_number"] = z
            z = z.type(torch.IntTensor).squeeze()
            f = torch.tensor(features[z]).type(torch.FloatTensor)
            if g.num_nodes() == 1:
                f = f.unsqueeze(0)
            g.ndata["atom_features"] = f

        self.prepare_batch = prepare_dgl_batch
        if line_graph:
            self.prepare_batch = prepare_line_graph_batch

            print("building line graphs")
            self.line_graphs = []
            for g in tqdm(graphs):
                lg = g.line_graph(shared=True)
                lg.apply_edges(compute_bond_cosines)
                self.line_graphs.append(lg)

        if classification:
            self.labels = self.labels.view(-1).long()
            print("Classification dataset.", self.labels)

    @staticmethod
    def _get_attribute_lookup(atom_features: str = "cgcnn"):
        """Build a lookup array indexed by atomic number."""
        max_z = max(v["Z"] for v in chem_data.values())

        # get feature shape (referencing Carbon)
        template = get_node_attributes("C", atom_features)

        features = np.zeros((1 + max_z, len(template)))

        for element, v in chem_data.items():
            z = v["Z"]
            x = get_node_attributes(element, atom_features)

            if x is not None:
                features[z, :] = x

        return features

    def __len__(self):
        """Get length."""
        return self.labels.shape[0]

    def __getitem__(self, idx):
        """Get StructureDataset sample."""
        g = self.graphs[idx]
        label = self.labels[idx]

        if self.transform:
            g = self.transform(g)

        if self.line_graph:
            return g, self.line_graphs[idx], label

        return g, label

    def setup_standardizer(self, ids):
        """Atom-wise feature standardization transform."""
        x = torch.cat(
            [
                g.ndata["atom_features"]
                for idx, g in enumerate(self.graphs)
                if idx in ids
            ]
        )
        self.atom_feature_mean = x.mean(0)
        self.atom_feature_std = x.std(0)

        self.transform = Standardize(
            self.atom_feature_mean, self.atom_feature_std
        )

    @staticmethod
    def collate(samples: List[Tuple[dgl.DGLGraph, torch.Tensor]]):
        """Dataloader helper to batch graphs cross `samples`."""
        graphs, labels = map(list, zip(*samples))
        batched_graph = dgl.batch(graphs)
        return batched_graph, torch.tensor(labels)

    @staticmethod
    def collate_line_graph(
        samples: List[Tuple[dgl.DGLGraph, dgl.DGLGraph, torch.Tensor]]
    ):
        """Dataloader helper to batch graphs cross `samples`."""
        graphs, line_graphs, labels = map(list, zip(*samples))
        batched_graph = dgl.batch(graphs)
        batched_line_graph = dgl.batch(line_graphs)
        if len(labels[0].size()) > 0:
            return batched_graph, batched_line_graph, torch.stack(labels)
        else:
            return batched_graph, batched_line_graph, torch.tensor(labels)


"""
if __name__ == "__main__":
    from jarvis.core.atoms import Atoms
    from jarvis.db.figshare import get_jid_data

    atoms = Atoms.from_dict(get_jid_data("JVASP-664")["atoms"])
    g = Graph.from_atoms(
        atoms=atoms,
        features="basic",
        get_prim=True,
        zero_diag=True,
        node_atomwise_angle_dist=True,
        node_atomwise_rdf=True,
    )
    g = Graph.from_atoms(
        atoms=atoms,
        features="cfid",
        get_prim=True,
        zero_diag=True,
        node_atomwise_angle_dist=True,
        node_atomwise_rdf=True,
    )
    g = Graph.from_atoms(
        atoms=atoms,
        features="atomic_number",
        get_prim=True,
        zero_diag=True,
        node_atomwise_angle_dist=True,
        node_atomwise_rdf=True,
    )
    g = Graph.from_atoms(atoms=atoms, features="basic")
    g = Graph.from_atoms(
        atoms=atoms, features=["Z", "atom_mass", "max_oxid_s"]
    )
    g = Graph.from_atoms(atoms=atoms, features="cfid")
    # print(g)
    d = g.to_dict()
    g = Graph.from_dict(d)
    num_nodes = g.num_nodes
    num_edges = g.num_edges
    print(num_nodes, num_edges)
    assert num_nodes == 48
    assert num_edges == 2304
    assert len(g.adjacency_matrix) == 2304
    # graph, color_map = get_networkx_graph(atoms)
    # nx.draw(graph, node_color=color_map, with_labels=True)
    # from jarvis.analysis.structure.neighbors import NeighborsAnalysis
"""



"""Module to simulate STM with Tershoff-Hamann approach."""
# Reference: https://www.nature.com/articles/s41597-021-00824-y
import matplotlib.pyplot as plt
from jarvis.io.vasp.outputs import Chgcar
import numpy as np
import matplotlib.transforms as mtransforms
import scipy
from jarvis.core.image import Image


class TersoffHamannSTM(object):
    """Generate constant height and constant current STM images."""

    def __init__(
        self,
        chg_name="PARCHG",
        min_size=50.0,
        height_tol=2,
        zcut=None,
        use_interpolated=True,
        crop_from_center=False,
        crop_mult=3,
        min_pixel=288,
        interp_step=1.0,
        skew=True,
        extend=0,
        ft_image_path=None,
        ft_image_zoom_factor=5,
        dpi=240,
        cmap="gray",
    ):
        """Initialize class with pathe of PARCHG and other input params."""
        # In original paper, extend used as 1
        chgcar = Chgcar(filename=chg_name)
        self.atoms = chgcar.atoms
        self.dim = chgcar.dim
        self.zmaxp = self.atoms.cart_coords[:, 2].max()
        self.nz = self.dim[2]
        volume = self.atoms.volume
        tmp = chgcar.chg[-1] * volume
        chg = tmp.reshape(self.dim[::-1]).T
        self.chg = chg
        self.cmap = cmap
        self.ft_image_path = ft_image_path
        self.ft_image_zoom_factor = ft_image_zoom_factor
        self.height_tol = height_tol
        self.a = self.atoms.lattice.a
        self.b = self.atoms.lattice.b
        self.c = self.atoms.lattice.c
        self.skew = skew
        self.zcut = zcut
        self.extend = extend
        z_frac_coords = self.atoms.frac_coords[:, 2]
        z_frac_coords_moved = []
        for i in z_frac_coords:
            if i > 0.5:
                i = i - 1
            elif i < -0.5:
                i = i + 1
            z_frac_coords_moved.append(i)
        self.crop_mult = crop_mult
        self.crop_from_center = crop_from_center
        if self.crop_from_center:
            min_size = min_size * self.crop_mult
        self.min_size = min_size
        self.zmaxp = max(np.array(z_frac_coords_moved) * self.c)
        rep_x = int(min_size / self.a) + self.extend
        rep_y = int(min_size / self.b) + self.extend
        self.repeat = [rep_x, rep_y]
        self.scell = self.atoms.make_supercell_matrix([rep_x, rep_y, 1])

        self.use_interpolated = use_interpolated
        self.interp_step = interp_step

        self.min_pixel = min_pixel
        self.dpi = dpi

    def constant_height(
        self,
        filename="testh.png",
    ):
        """Get iso-height image."""
        tol = self.height_tol
        if not self.zcut:
            self.zcut = int((self.zmaxp + tol) / self.c * self.nz)
        # print("zcut", self.zcut, self.repeat)
        info = {}

        img_ext = np.tile(self.chg[:, :, self.zcut], self.repeat)
        if not self.use_interpolated:
            exts = (
                0,
                self.a * self.repeat[0],
                0,
                self.b * (self.repeat[1] - 1),
            )
            plt.close()
            fig, ax = plt.subplots()
            plt.xticks([])
            plt.yticks([])
            if self.skew:
                tmp = 90 - self.atoms.lattice.angles[2]
            else:
                tmp = 0
            data = self.get_plot(
                ax,
                img_ext,
                exts,
                mtransforms.Affine2D().skew_deg(tmp, 0),
            )
            info["data"] = data
            info["img_ext"] = img_ext
            fig.subplots_adjust(bottom=0, top=1, left=0.0, right=1)
            plt.savefig(
                filename, dpi=self.dpi
            )  # , bbox_inches="tight", pad_inches=0.0, dpi=240)
            plt.close()
        else:
            img_ext = self.get_interpolated_data(img_data=img_ext)
            # print ('img_ext',img_ext)
            plt.close()
            # fig, ax = plt.subplots()
            plt.imshow(img_ext, interpolation="none", cmap=self.cmap)
            # ax.set_aspect('equal')
            plt.axis("off")
            plt.savefig(filename, dpi=self.dpi)

            plt.close()

            plt.gca().set_axis_off()
            plt.subplots_adjust(
                top=1, bottom=0, right=1, left=0, hspace=0, wspace=0
            )
            plt.margins(0, 0)
            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())
            # print ('img_ext',img_ext)
            plt.close()

        info["scell"] = self.scell
        info["zcut"] = self.zcut
        info["img_ext"] = img_ext
        if self.crop_from_center:
            im = Image.from_file(filename)

            p_x = int(self.min_pixel / self.crop_mult)
            # print('shape',im.values.shape,p_x)
            im = Image.crop_from_center(
                image_path=filename, target_left=p_x, target_right=p_x
            )
            info["img_ext"] = im
            plt.imshow(im, cmap=self.cmap)
            plt.tight_layout()
            plt.axis("off")
            plt.savefig(filename, dpi=self.dpi)
            plt.close()
        if self.ft_image_path is not None:
            im = Image.from_file(filename).fourier_transform2D(
                zoom_factor=self.ft_image_zoom_factor
            )
            im.save(self.ft_image_path)
        return info

    def constant_current(
        self,
        pc=None,
        ext=0.15,
        filename="testc.png",
        use_interpolated=True,
    ):
        """Return the constant-current cut the charge density."""
        zmax_ind = int(self.zmaxp / self.c * self.nz) + 1
        info = {}
        tol = self.height_tol
        # Find what z value is near the current, and take avergae
        if not self.zcut:
            self.zcut = int((self.zmaxp + tol) / self.c * self.nz)
        zext = int(self.nz * ext)
        zcut_min = self.zcut - zext
        if zcut_min < zmax_ind:
            zcut_min = zmax_ind
        zcut_max = self.zcut + zext
        if pc is None:
            c = np.average(self.chg[:, :, self.zcut])
        else:
            tmp = np.zeros(zcut_max - zcut_min)
            for ii in range(tmp.size):
                tmp[ii] = np.average(self.chg[:, :, zcut_min + ii])
            c = np.linspace(tmp.min(), tmp.max(), 100)[pc]

        # height of iso-current
        img = np.argmin(np.abs(self.chg[:, :, zcut_min:zcut_max] - c), axis=2)
        img_ext = np.tile(img, self.repeat[::-1]) + self.zcut - zext
        if not use_interpolated:
            fig, ax = plt.subplots()
            exts = (
                0,
                self.a * self.repeat[0],
                0,
                self.b * (self.repeat[1] - 1),
            )
            plt.xticks([])
            plt.yticks([])
            if self.skew:
                tmp = 90 - self.atoms.lattice.angles[2]
            else:
                tmp = 0

            data = self.get_plot(
                ax,
                img_ext,
                exts,
                mtransforms.Affine2D().skew_deg(tmp, 0),
            )
            info["data"] = data
            info["img_ext"] = img_ext
            plt.savefig(filename, dpi=self.dpi)
            plt.close()
        else:
            img_ext = self.get_interpolated_data(img_data=img_ext)
            plt.close()
            # fig, ax = plt.subplots()
            plt.imshow(img_ext, interpolation="none", cmap=self.cmap)
            # ax.set_aspect('equal')
            plt.axis("off")
            plt.savefig(filename, dpi=self.dpi)
            plt.close()
        if self.crop_from_center:
            im = Image.from_file(filename)

            p_x = int(self.min_pixel / self.crop_mult)
            # print('shape',im.values.shape,p_x)
            im = Image.crop_from_center(
                image_path=filename, target_left=p_x, target_right=p_x
            )
            plt.imshow(im, cmap=self.cmap)
            plt.axis("off")
            plt.savefig(filename, dpi=self.dpi)
            plt.close()

        info["scell"] = self.scell
        info["zcut"] = self.zcut
        return info

    def get_plot(self, ax, Z, extent, transform):
        """Apply affine transformation."""
        ax.axis("off")
        im = ax.imshow(
            Z,
            interpolation="none",
            origin="lower",
            extent=extent,
            clip_on=True,
            aspect="equal",
        )  # ,cmap=plt.get_cmap('gray')

        trans_data = transform + ax.transData
        im.set_transform(trans_data)

        # display intended extent of the image
        x1, x2, y1, y2 = im.get_extent()
        data = ax.plot(
            [x1, x2, x2, x1, x1],
            [y1, y1, y2, y2, y1],
            "y--",
            transform=trans_data,
        )
        print("min Z and maxZ", np.min(Z), np.max(Z))
        return data

    def get_interpolated_data(self, img_data=[]):
        """Get interpolated data after moving pixels."""
        x = []
        y = []
        zz = []
        xy = []
        step = self.interp_step
        atoms = self.atoms
        for i in range(img_data.shape[0]):
            for j in range(img_data.shape[1]):
                z = img_data[i][j]
                xyz = i * atoms.lattice_mat[0] + j * atoms.lattice_mat[1]
                x.append(xyz[0])
                y.append(xyz[1])
                zz.append(z)
                xy.append([xyz[0], xyz[1]])

        grid_x, grid_y = np.mgrid[
            min(x) : max(x) : step, min(y) : max(y) : step
        ]
        # stepx=(max(x)-min(x))/bins
        # stepy=(max(y)-min(y))/bins
        # grid_x, grid_y = np.mgrid[min(x):max(x):stepx, min(y):max(y):stepy]
        interp = scipy.interpolate.griddata(xy, zz, (grid_x, grid_y)).T
        # plt.scatter(x, y, 1, zz)
        # plt.savefig("ok.png")
        # plt.close()
        return interp


"""
if __name__ == "__main__":
    plt.switch_backend("agg")
    f = "PARCHG"
    t = TersoffHamannSTM(chg_name=f).constant_height()
    t = TersoffHamannSTM(chg_name=f).constant_current()
"""



"""Module to run QE jobs."""

from jarvis.io.qe.inputs import QEinfile
from jarvis.io.qe.outputs import QEout
import os
from jarvis.db.jsonutils import loadjson, dumpjson
import subprocess
from jarvis.core.atoms import Atoms
from jarvis.core.kpoints import Kpoints3D


class QEjob(object):
    """Module to run generic QE job."""

    def __init__(
        self,
        atoms=None,
        kpoints=None,
        input_params={},
        qe_cmd="pw.x",
        jobname="test",
        psp_dir=None,
        url=None,
        output_file="qe.out",
        input_file="qe.in",
        stderr_file="std.err",
        psp_temp_name=None,
    ):
        """Intitialize class."""
        self.atoms = atoms
        self.kpoints = kpoints
        self.input_params = input_params
        self.qe_cmd = qe_cmd
        self.jobname = jobname
        self.psp_dir = psp_dir
        self.url = url
        self.psp_temp_name = psp_temp_name
        self.input_file = input_file
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.qeinput = QEinfile(
            atoms=self.atoms,
            kpoints=self.kpoints,
            psp_dir=self.psp_dir,
            input_params=self.input_params,
            url=self.url,
            psp_temp_name=self.psp_temp_name,
        )

    def write_input(self):
        """Write inputs."""
        self.qeinput.write_file(self.input_file)

    def to_dict(self):
        """Get dictionary."""
        info = {}
        info["atoms"] = self.atoms.to_dict()

        info["kpoints"] = self.kpoints.to_dict()
        info["qe_cmd"] = self.qe_cmd
        info["psp_dir"] = self.psp_dir
        info["url"] = self.url
        return info

    @classmethod
    def from_dict(self, info={}):
        """Load from a dictionary."""
        return QEjob(
            atoms=Atoms.from_dict(info["atoms"]),
            kpoints=Kpoints3D.from_dict(info["kpoints"]),
            qe_cmd=info["qe_cmd"],
            psp_dir=info["psp_dir"],
            url=info["url"],
        )

    def runjob(self):
        """Run job and make or return a metadata file."""
        fname = self.jobname + ".json"
        info = {}
        if not os.path.exists(fname):
            self.write_input()
            self.run()
            info = self.parse_outputs()
            if info["job_done"]:
                dumpjson(data=info, filename=fname)
        else:
            info = loadjson(fname)
        return info

    def run(self):
        """Use subprocess to tun a job."""
        with open(self.output_file, "w") as f_std, open(
            self.stderr_file, "w", buffering=1
        ) as f_err:
            # use line buffering for stderr
            cmd = self.qe_cmd + "<" + self.input_file
            print("cmd", cmd)
            p = subprocess.Popen(cmd, shell=True, stdout=f_std, stderr=f_err)
            p.wait()
        return p

    def parse_outputs(self):
        """Parse outputs from .out or .xml files."""
        info = {}
        info["out_path"] = "na"
        info["xml_path_"] = "na"
        info["total_energy"] = "na"
        info["job_done"] = "na"
        out_path = os.path.abspath(self.output_file)
        if os.path.exists(out_path):
            try:
                info["out_path"] = out_path
                qe_out = QEout(filename=out_path)
                job_done = qe_out.job_done
                info["job_done"] = job_done
                total_energy = qe_out.get_total_energy()
                info["total_energy"] = total_energy
            except Exception as exp:
                print("Exception", exp)
                pass
        if (
            "control" in self.input_params
            and "prefix" in self.input_params["control"]
        ):
            print("HERE1")
            xml_path = os.path.abspath(
                os.path.join(
                    self.input_params["control"]["prefix"]
                    .strip('"')
                    .strip("'")
                    + ".save",
                    "data-file-schema.xml",
                )
            )
            print("HERE2", xml_path)
            if os.path.exists(xml_path):
                info["xml_path"] = xml_path
        return info



"""Version number."""

__version__ = "2024.4.10"

import os


def test(*args):
    """Run pytest in the base of jarvis."""
    import pytest

    path = os.path.join(os.path.split(__file__)[0], "tests")
    pytest.main(args=[path] + list(args))



"""Module to simulate STEM images using convoltuin approximation."""
# Adapted from https://github.com/jacobjma/fourier-scale-calibration
import numpy as np
from scipy.interpolate import interp1d

# from numbers import Number
from jarvis.core.utils import gaussian
from jarvis.core.utils import lorentzian2 as lorentzian
from jarvis.core.atoms import Atoms  # , get_supercell_dims, crop_square
from typing import List


class STEMConv(object):
    """Module to simulate STEM images using convoltuin approximation."""

    def __init__(
        self,
        atoms=None,
        output_size=[50, 50],
        power_factor=1.7,
        gaussian_width=0.5,
        lorentzian_width=0.5,
        intensity_ratio=0.5,
        nbins=100,
        tol=0.5,
        crop=False,
    ):
        """Intitialize the class."""
        self.atoms = atoms
        self.output_size = output_size
        self.power_factor = power_factor
        self.gaussian_width = gaussian_width
        self.lorentzian_width = lorentzian_width
        self.intensity_ratio = intensity_ratio
        self.nbins = nbins
        self.tol = tol
        self.crop = crop

    def superpose_deltas(self, positions, array):
        """Superpose deltas."""
        z = 0
        shape = array.shape[-2:]
        rounded = np.floor(positions).astype(np.int32)
        rows, cols = rounded[:, 0], rounded[:, 1]

        array[z, rows, cols] += (1 - (positions[:, 0] - rows)) * (
            1 - (positions[:, 1] - cols)
        )
        array[z, (rows + 1) % shape[0], cols] += (positions[:, 0] - rows) * (
            1 - (positions[:, 1] - cols)
        )
        array[z, rows, (cols + 1) % shape[1]] += (
            1 - (positions[:, 0] - rows)
        ) * (positions[:, 1] - cols)
        array[z, (rows + 1) % shape[0], (cols + 1) % shape[1]] += (
            rows - positions[:, 0]
        ) * (cols - positions[:, 1])
        return array

    def simulate_surface(
        self,
        atoms: Atoms,
        px_scale: float = 0.2,
        eps: float = 0.6,
        rot: float = 0,
        shift: List = [0, 0],
    ):
        """Simulate a STEM image.

        atoms: jarvis.core.Atoms material slab
        px_scale: pixel size in angstroms/px
        eps: tolerance factor (angstroms)
        for rendering atoms outside the field of view
        rot: rotation about the image center (degrees)
        shift: rigid translation of field of view [dx, dy] (angstroms)

        """
        shift = np.squeeze(shift)
        output_px = np.squeeze(self.output_size)  # px

        # field of view size in angstroms
        view_size = px_scale * (output_px - 1)

        # construct a supercell grid big enough to fill the field of view
        cell_extent = atoms.lattice.abc[0:2]  # np.diag(atoms.lattice_mat)[:2]

        cells = ((view_size // cell_extent) + 1).astype(int)
        # print ('cells',cells)
        atoms = atoms.make_supercell_matrix((3 * cells[0], 3 * cells[1], 1))

        # Set up real-space grid (in angstroms)
        # construct the probe array with the output target size
        # fftshift, pad, un-fftshift
        x = np.fft.fftfreq(output_px[0]) * output_px[0] * px_scale
        y = np.fft.fftfreq(output_px[1]) * output_px[1] * px_scale
        r = np.sqrt(x[:, None] ** 2 + y[None] ** 2)

        # construct the probe profile centered
        # at (0,0) on the periodic spatial grid
        x = np.linspace(0, 4 * self.lorentzian_width, self.nbins)
        profile = gaussian(
            x, self.gaussian_width
        ) + self.intensity_ratio * lorentzian(x, self.lorentzian_width)
        profile /= profile.max()
        f = interp1d(x, profile, fill_value=0, bounds_error=False)
        intensity = f(r)

        # shift the probe profile to the center
        # apply zero-padding, and shift back to the origin
        margin = int(np.ceil(5 / px_scale))  # int like 20
        intensity = np.fft.fftshift(intensity)
        intensity = np.pad(intensity, (margin, margin))
        intensity = np.fft.fftshift(intensity)

        # project atomic coordinates onto the image
        # center them as well
        centroid = np.mean(atoms.cart_coords[:, :2], axis=0)

        # center atom positions around (0,0)
        pos = atoms.cart_coords[:, :2] - centroid

        # apply field of view rotation
        # (actually rotate the lattice coordinates)
        if rot != 0:
            rot = np.radians(rot)
            R = np.array(
                [[np.cos(rot), -np.sin(rot)], [np.sin(rot), np.cos(rot)]]
            )
            pos = pos @ R

        # shift to center of image
        pos += view_size / 2

        # apply rigid translation of atoms wrt image field of view
        pos += shift

        # select only atoms in field of view
        in_view = (
            (pos[:, 0] > -eps)
            & (pos[:, 0] < view_size[0] + eps)
            & (pos[:, 1] > -eps)
            & (pos[:, 1] < view_size[1] + eps)
        )
        # pos = pos[in_view]
        numbers = np.array(atoms.atomic_numbers)
        # numbers = numbers[in_view]

        atom_px = pos / px_scale  # AA / (AA/px) -> px

        # atom_px = atom_px + margin

        render = in_view
        # render = (
        #     (pos[:, 0] > 0)
        #     & (pos[:, 0] < view_size[0])
        #     & (pos[:, 1] > 0)
        #     & (pos[:, 1] < view_size[1])
        # )

        numbers_render = numbers[render]
        # # shift atomic positions to offset zero padding
        atom_px_render = atom_px[render] + margin

        # initialize arrays with zero padding
        array = np.zeros((1,) + intensity.shape)  # adding extra 1
        mask = np.zeros((1,) + intensity.shape)
        # print(f"intensity: {array.shape}")
        for number in np.unique(np.array(atoms.atomic_numbers)):

            temp = np.zeros((1,) + intensity.shape)
            temp = self.superpose_deltas(
                atom_px_render[numbers_render == number], temp
            )
            array += temp * number ** self.power_factor
            temp = np.where(temp > 0, number, temp)
            mask += temp[0]

        # FFT convolution of beam profile and atom position delta functions
        array = np.fft.ifft2(np.fft.fft2(array) * np.fft.fft2(intensity)).real

        # crop the FFT padding and fix atom coordinates relative to
        # the image field of view
        sel = slice(margin, -margin)
        array = array[0, sel, sel]
        mask = mask[0, sel, sel]
        # atom_px = atom_px - margin

        atom_px = pos[in_view] / px_scale
        numbers = numbers[in_view]

        return array, mask, atom_px, numbers


"""
if __name__ == "__main__":
    from jarvis.core.atoms import crop_squre
    from jarvis.db.figshare import data, get_jid_data
    import matplotlib.pyplot as plt
    from jarvis.core.atoms import Atoms, ase_to_atoms, get_supercell_dims

    plt.switch_backend("agg")

    a = Atoms.from_dict(get_jid_data("JVASP-667")["atoms"])
    stem = STEMConv(output_size=(256, 256))
    p, mask, atom_x, nb = stem.simulate_surface(
        a, px_scale=0.15, eps=0.6, rot=2, shift=[-0.2, 0.3]
    )
    plt.imshow(p, origin="lower", cmap="plasma")
    plt.scatter(atom_x[:, 1], atom_x[:, 0], marker="o",
    facecolors="none", edgecolors="r"
    )
    plt.savefig("stem_example.png")
    plt.close()
"""



"""Modules for making point-defect substituions."""
from jarvis.analysis.structure.spacegroup import Spacegroup3D
from jarvis.core.utils import rand_select
from jarvis.core.atoms import Atoms, get_supercell_dims


def generate_defect(
    atoms=None,
    enforce_c_size=10.0,
    on_conventional_cell=False,
    extend=1,
    subs_element="Al",
    selected_element=None,
):
    """Provide function to generate substitution defects."""
    if on_conventional_cell:
        atoms = Spacegroup3D(atoms).conventional_standard_structure
    if enforce_c_size is not None:
        dims = get_supercell_dims(
            atoms, enforce_c_size=enforce_c_size, extend=extend
        )
        supercell_size = [dims[0], dims[1], dims[2]]
        # atoms = atoms.make_supercell(supercell_size)
    spg = Spacegroup3D(atoms)
    wyckoffs = spg._dataset["wyckoffs"]
    atoms.props = wyckoffs
    supercell = atoms.make_supercell(supercell_size)
    props = rand_select(supercell.props)
    subs = []
    # print(props)
    for i, j in props.items():
        info = {}
        elements = supercell.elements.copy()
        if selected_element is not None:
            if elements[j] == selected_element:

                elements[j] = subs_element

                a = Atoms(
                    lattice_mat=supercell.lattice_mat,
                    coords=supercell.coords,
                    cartesian=supercell.cartesian,
                    elements=elements,
                )
                info["props"] = props
                info["atoms"] = atoms.to_dict()
                info["defect_atoms"] = a.to_dict()
                info["supercell_size"] = list(supercell_size)
                # print(a.elements)
                subs.append(info)
        else:
            elements[j] = subs_element
            a = Atoms(
                lattice_mat=supercell.lattice_mat,
                coords=supercell.coords,
                cartesian=supercell.cartesian,
                elements=elements,
            )
            info["props"] = props
            info["atoms"] = atoms.to_dict()
            info["defect_atoms"] = a.to_dict()
            info["supercell_size"] = list(supercell_size)
            # print(a.elements)
            subs.append(info)
    return subs


"""
x = generate_defect(atoms=a, selected_element="Br")
"""



"""
Code to predict properties and their uncertainty.

ML model used: lgbm
"""
# Ref: https://doi.org/10.1021/acsomega.1c03752
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from jarvis.ai.pkgs.utils import regr_scores
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF
from sklearn.gaussian_process.kernels import WhiteKernel
from sklearn.gaussian_process.kernels import RationalQuadratic

# from sklearn.gaussian_process.kernels import ExpSineSquared
# from sklearn.gaussian_process.kernels import ConstantKernel as C
# import joblib
from joblib import dump
from collections import OrderedDict


def GaussianProcesses(
    x,
    y,
    jid,
    cv=2,
    n_jobs=-1,
    n_iter=10,
    random_state=508842607,
    scoring="neg_mean_absolute_error",
    prop="exfoliation_energy",
    write_files=True,
):
    """
    Perform Gaussian Processes and determine prediction intervals.

    NOTE: it can deal with ~50 descriptors AT MAX ==> reduction of
          the CFID descriptors is needed
    """
    # TODO: Make writing file in proper python format

    # STEP-2: Splitting the data
    # ***************************
    # 90-10% split for train test

    X_train, X_test, y_train, y_test, jid_train, jid_test = train_test_split(
        x, y, jid, random_state=1, test_size=0.1
    )
    # print ('lenx len y',len(x[0]),len(y))

    # STEP-3: Use a specific ML modeli: GP in this case
    # *************************************************
    #
    # Model definition
    # ================
    # The Kernel can be chanded depending on the property under examination.
    # In this example we use a composite kernel that includes long range
    # effects (RBF), medium term irregularities (RationalQuadratic), and
    # white noise
    # Periodic trends can be added using ExpSineSquared
    # Note: RBF has one parameter for each descriptor used ==> if using a
    #       different number of descriptors, the number of parameters has
    #       to be changed
    RQ1 = 316**2 * RationalQuadratic(
        alpha=0.067,
        length_scale=43.5,
        length_scale_bounds=(1e0, 1e2),
        alpha_bounds=(1e-3, 1e1),
    )
    RQ2 = 316**2 * RationalQuadratic(
        alpha=0.00473,
        length_scale=0.0796,
        length_scale_bounds=(1e-2, 1e0),
        alpha_bounds=(1e-3, 1e1),
    )
    RBF1 = 316**2 * RBF(
        length_scale=[
            16,
            1.48e03,
            9.92e04,
            0.858,
            1.02,
            2.58,
            2.72e03,
            2.22e03,
            0.839,
            0.5,
            6.15,
            130,
            3.9e03,
            3.1,
            0.675,
            6.04,
            1.18e03,
            2.9,
            4.56e03,
            333,
            0.0167,
            797,
            2.37e03,
            1.81e03,
        ]
    )
    White = WhiteKernel(noise_level=0.7, noise_level_bounds=(0.05, 0.8))
    kernel = RQ1 + RQ2 + RBF1 + White

    mid_model = GaussianProcessRegressor(
        kernel=kernel,
        alpha=0.0001,
        n_restarts_optimizer=3,
        normalize_y=True,
        random_state=int(774057201),
    )

    # Model fitting
    # =============
    scaler = StandardScaler().fit(X_train)
    scaler.transform(X_train)
    scaler.transform(X_test)

    mid_model.fit(X_train, y_train)

    dump(mid_model, "model.joblib")
    print("")
    print("Params:")
    print(mid_model.get_params(deep=True))
    print("Initial: ", kernel)
    print("Optimum: ", mid_model.kernel_)
    print(
        "Log-Marginal-Likelihood: ",
        mid_model.log_marginal_likelihood(mid_model.kernel_.theta),
    )
    print("")
    print("")

    # PREDICTIONS and UQ
    # ==================
    pred, sigma = mid_model.predict(X_test, return_std=True)
    print("Model       mae     rmse")
    reg_sc = regr_scores(y_test, pred)
    info = OrderedDict()
    info["MAE_Mid"] = reg_sc["mae"]
    actual = y_test
    print("Mid:", round(reg_sc["mae"], 3), round(reg_sc["rmse"], 3))

    fout2 = open("Intervals1.dat", "w")
    line0 = "#    Jid      Observed       pred_Lower       pred_Mid    "
    line1 = line0 + "RealErr(Mid)     "
    line = line1 + "pred_Upper    PredError    Pred_inBounds\n"
    fout2.write(line)
    sum = 0.0
    count = 0
    MAE_err = 0.0
    for ii in range(len(y_test)):
        ss = float(sigma[ii])
        pp = float(pred[ii])
        sum = sum + ss
        lower = pp - ss
        upper = pp + ss
        err = ss
        real_err = float(abs(y_test[ii] - pp))
        err_err = abs(real_err - err)
        MAE_err = MAE_err + err_err
        if (pp >= lower) and (pp <= upper):
            in_bounds = "True"
            count = count + 1
        else:
            in_bounds = "False"
        line2_0 = str(ii) + " " + jid[ii] + " " + str(y_test[ii]) + " "
        line2_1 = line2_0 + str(lower) + " " + str(pp) + " " + str(real_err)
        line2 = (
            line2_1
            + " "
            + str(upper)
            + " "
            + str(ss)
            + " "
            + str(in_bounds)
            + "\n"
        )
        fout2.write(line2)
    print("")
    print("Number of test materials= " + str(len(actual)))
    print(
        "Percentage of in-bound results= "
        + str((float(count) / (len(actual))) * 100)
        + "%"
    )
    print(" ")
    MAE_error = float(MAE_err) / (len(actual))
    print("MAE predicted error = " + str(MAE_error))
    info["MAE_Error"] = MAE_error
    return info



from jarvis.analysis.thermodynamics.energetics import (
    form_enp,
    get_twod_defect_energy,
    PhaseDiagram,
    jid_hull
)
from jarvis.db.figshare import data
import os
from jarvis.io.vasp.outputs import Vasprun
from jarvis.core.composition import Composition

tmp_xml = os.path.join(os.path.dirname(__file__), "JVASP-667_C_C_c.xml")
vrun = Vasprun(tmp_xml)

dft_3d=data('dft_3d')

def test_get_twod_defect_energy():
    Ef = get_twod_defect_energy(vrun=vrun, jid="JVASP-667", atom="C")
    print(Ef)


def test_form_enp():
    atoms = vrun.all_structures[-1]
    total_energy = vrun.final_energy
    Ef = form_enp(atoms=atoms, total_energy=total_energy)
    print(Ef)


def test_phasediag():
    system = ["Al", "O"]
    system = ["Bi", "Se"]
    system = ["Cu", "Au"]
    system = ["Al", "Ni"]
    system = ["Ga", "Al", "N"]
    system = ["Ni", "Al", "O"]
    system = ["O", "Al", "Ga", "N"]
    system = ["Ni", "Fe", "Cr"]
    system=['Mo','Se']
    x = []
    y = []
    for i in dft_3d:
        formula = i["formula"]
        comp = Composition.from_string(formula)
        atom_frac = comp.atomic_fraction
        all_elms = list(comp.to_dict())
        if (set(all_elms)).issubset(set(system)):  # and len(set(all_elms))!=1:
            x.append([i["formula"], i["formation_energy_peratom"], i["jid"]])

    pd = PhaseDiagram(
        x, verbose=True, only_plot_stable=True, only_label_stable=True
    )

    pd.plot()
def test_jid_hull():
   x=jid_hull(jid='JVASP-114723',dataset=dft_3d)
# test_get_twod_defect_energy()
# test_form_enp()



"""Module to generate XML file for QE calculation."""
import gzip
import os
import xmltodict
from jarvis.core.atoms import get_supercell_dims
from jarvis.analysis.structure.spacegroup import Spacegroup3D
from jarvis.core.utils import stringdict_to_xml, xml_to_dict, array_to_string
from jarvis.io.qe.outputs import ProjHamXml, DataFileSchema
from jarvis.analysis.thermodynamics.energetics import (
    form_enp,
    get_unary_qe_tb_energy,
)

# import pprint


def parse_material_calculation_folder(
    path="Ag_Al_kspace/dimer.in_vnscf_coords_min_2",
    source="JARVIS-DFT-QE",
):
    """Parse QE calculation folder."""
    save_path = os.path.join(path, "qe.save")
    info = {}
    if not os.path.exists(save_path):
        raise ValueError("Doesnt exist:", save_path)
    jid_path = os.path.join(path, "JARVIS-ID")
    keys = [
        "jid",
        "source_folder",
        "final_energy",
        "final_energy_breakdown",
        # "initial_structure",
        "natoms",
        "final_spacegroup_number",
        "final_spacegroup_symbol",
        "final_pointgroup_symbol",
        "final_crystal_system",
        # "initial_spacegroup_number",
        # "initial_spacegroup_symbol",
        # "initial_pointgroup_symbol",
        # "initial_crystal_system",
        "energy_per_atom",
        "elements",
        "formula",
        "number_uniq_species",
        "final_structure",
        "efermi",
        "functional",
        "indir_gap",
        "nkpts",
        "qe_version",
        "is_spin_polarized",
        "is_spin_orbit",
        "data_source",
        "forces",
        "stress",
        "nelec",
        "f_enp",
        "dos",
    ]
    for i in keys:
        info[i] = "na"
    if os.path.exists(jid_path):
        f = open(jid_path, "r")
        jid = f.read().splitlines()[0]
        f.close()

        info["jid"] = jid
        info["tmp_id"] = "'" + jid + "'"
        info["source_folder"] = "'" + path + "'"

    for i in os.listdir(save_path):
        if "data-file-schema" in i:
            if ".gz" in i:
                dat_path = os.path.join(save_path, i)
                f = gzip.open(dat_path, "rb")
                file_content = f.read()
                data = xmltodict.parse(file_content)
                if "output" in data["qes:espresso"]:
                    set_key = "output"
                elif "step" in data["qes:espresso"]:
                    set_key = "step"
                else:
                    raise ValueError("Inconsisten QE version.")
                data_schm = DataFileSchema(data=data, set_key=set_key)
            else:
                dat_path = os.path.join(save_path, i)
                data = xml_to_dict(dat_path)
                if "output" in data["qes:espresso"]:
                    set_key = "output"
                elif "step" in data["qes:espresso"]:
                    set_key = "step"
                else:
                    raise ValueError("Inconsisten QE version.")
                data_schm = DataFileSchema(data=data, set_key=set_key)
            info["final_energy"] = data_schm.final_energy
            for ii, jj in data_schm.final_energy_breakdown.items():
                info[ii] = jj
            final_strt = data_schm.final_structure
            info["final_structure"] = (
                "'"
                + data_schm.final_structure.get_string().replace("\n", "\\n")
                + "'"
            )
            try:
                print("data_schm.final_energy", data_schm.final_energy)
                f_enp = form_enp(
                    atoms=final_strt,
                    chem_pots=get_unary_qe_tb_energy(),
                    total_energy=data_schm.final_energy,
                )
                info["f_enp"] = round(f_enp, 4)
            except Exception as exp:
                print(exp)
                pass
            dim = get_supercell_dims(final_strt)
            info["xyz"] = (
                '"'
                + str(
                    final_strt.make_supercell_matrix(dim).get_xyz_string
                ).replace("\n", "\\n")
                + '"'
            )
            info["natoms"] = final_strt.num_atoms
            final_spg = Spacegroup3D(final_strt)
            info["final_spacegroup_number"] = final_spg.space_group_number
            info["final_spacegroup_symbol"] = final_spg.space_group_symbol
            info["final_pointgroup_symbol"] = final_spg.point_group_symbol
            info["final_crystal_system"] = final_spg.crystal_system
            final_lat_params = final_strt.lattice.parameters
            info["final_a"] = round(final_lat_params[0], 2)
            info["final_b"] = round(final_lat_params[1], 2)
            info["final_c"] = round(final_lat_params[2], 2)
            info["final_alpha"] = round(final_lat_params[3], 2)
            info["final_beta"] = round(final_lat_params[4], 2)
            info["final_gamma"] = round(final_lat_params[5], 2)
            info["final_density"] = round(final_strt.density, 3)

            initial_strt = data_schm.initial_structure
            initial_lat_params = initial_strt.lattice.parameters
            info["initial_a"] = round(initial_lat_params[0], 2)
            info["initial_b"] = round(initial_lat_params[1], 2)
            info["initial_c"] = round(initial_lat_params[2], 2)
            info["initial_alpha"] = round(initial_lat_params[3], 2)
            info["initial_beta"] = round(initial_lat_params[4], 2)
            info["initial_gamma"] = round(initial_lat_params[5], 2)
            info["initial_density"] = round(initial_strt.density, 3)
            """
            initial_spg = Spacegroup3D(initial_strt)
            info["initial_spacegroup_number"] = initial_spg.space_group_number
            info["initial_spacegroup_symbol"] = initial_spg.space_group_symbol
            info["initial_pointgroup_symbol"] = initial_spg.point_group_symbol
            info["initial_crystal_system"] = initial_spg.crystal_system
            info["initial_structure"] = (
                "'"
                + data_schm.initial_structure.get_string().replace("\n", "\\n")
                + "'"
            )
            """

            info["energy_per_atom"] = round(
                float(data_schm.energy_per_atom), 4
            )

            info["cif"] = ""
            try:
                cif_filename = os.path.join(os.getcwd(), "atoms.cif")
                final_strt.write_cif(filename="atoms.cif")
                cif_f = open(cif_filename, "r")
                cif_lines = cif_f.read()  # .splitlines()
                cif_f.close()
                os.remove(cif_filename)
                info["cif"] = '"' + str(cif_lines).replace("\n", "\\n") + '"'
            except Exception as exp:
                print("CIF exception", exp)
                pass
            info["elements"] = ",".join(final_strt.uniq_species)
            info["tmp_elements"] = (
                "'" + ",".join(final_strt.uniq_species) + "'"
            )
            info["formula"] = final_strt.composition.reduced_formula
            info["tmp_formula"] = (
                "'" + final_strt.composition.reduced_formula + "'"
            )
            info["number_uniq_species"] = len(final_strt.uniq_species)
            info["efermi"] = round(data_schm.efermi, 4)
            info["functional"] = data_schm.functional
            info["indir_gap"] = round(data_schm.indir_gap, 3)
            info["nkpts"] = data_schm.nkpts
            info["nelec"] = data_schm.nelec
            info["qe_version"] = data_schm.qe_version
            info["is_spin_polarized"] = data_schm.is_spin_polarized
            info["is_spin_orbit"] = data_schm.is_spin_orbit
            info["data_source"] = source
            info["forces"] = (
                "'"
                + array_to_string(
                    [" ".join(map(str, j)) for j in data_schm.forces]
                )
                + "'"
            )
            info["stress"] = (
                "'"
                + array_to_string(
                    [" ".join(map(str, j)) for j in data_schm.stress]
                )
                + "'"
            )
    try:
        projham_xml_path_gz = os.path.join(path, "projham_K.xml.gz")
        projham_xml_path = os.path.join(path, "projham_K.xml")
        if os.path.exists(projham_xml_path):
            energies, dos, pdos, names = ProjHamXml(
                filename=projham_xml_path
            ).dos()
            line = (
                "<edos_energies>'"
                + ",".join(map(str, energies))
                + "'</edos_energies>"
            )
            line += (
                "<total_edos_up>'"
                + ",".join(map(str, dos))
                + "'</total_edos_up>"
            )
            line += "<elemental_dos>"
            for n, nn in enumerate(names):
                pdos_n = pdos[:, n]
                line += str(nn) + "_" + ",".join(map(str, pdos_n)) + ";"
            line += "</elemental_dos>"

            info["dos"] = line
        elif os.path.exists(projham_xml_path_gz):
            energies, dos, pdos, names = ProjHamXml(
                filename=projham_xml_path_gz
            ).dos()
            line = (
                "<edos_energies>'"
                + ",".join(map(str, energies))
                + "'</edos_energies>"
            )
            line += (
                "<total_edos_up>'"
                + ",".join(map(str, dos))
                + "'</total_edos_up>"
            )
            line += "<elemental_dos>'"
            for n, nn in enumerate(names):
                pdos_n = pdos[:, n]
                line += str(nn) + "_" + ",".join(map(str, pdos_n)) + ";"
            line += "'</elemental_dos>"
            info["dos"] = line
        else:
            info["dos"] = "na"
    # print(pprint.pprint(info))
    except Exception as exp:
        print(exp)
        pass
    return info


def write_xml(path="", filename="temp.xml"):
    """Write XML file."""
    basic = parse_material_calculation_folder(path=path)
    f = open(filename, "w")
    basic_xml = stringdict_to_xml(basic)
    line = ""
    line = '<?xml version="1.0" encoding="UTF-8"?>\n'
    line += '<?xml-stylesheet type="text/xsl" '
    line += 'href="jarvisqe.xsl"?>\n'

    line += "<basic_info>"

    line += basic_xml
    line += "</basic_info>"
    f.write(line)
    f.close()


"""
#path='/rk2/knc6/UniveralTB/julia_data/TEST_ELS/POSCAR_JVASP-1002'
#path='/rk2/knc6/UniveralTB/julia_data/TEST_BIN/POSCAR_JVASP-10016'
if __name__ == "__main__":
    #path = "/rk2/knc6/UniveralTB/julia_data/
    #Si_C_kspace/POSCAR_tio2_rutile_vnscf_vol_1"
    write_xml(path=path)
"""



from jarvis.core.utils import (
    update_dict,
    recast_array_on_uniq_array_elements,
    stringdict_to_xml,
    array_to_string,
    check_url_exists,
)

from jarvis.ai.pkgs.utils import mean_absolute_deviation


def test_utils():
    info = {"a": "b", "c": "d"}
    sd = stringdict_to_xml(info)
    sd = stringdict_to_xml(info, enforce_string=True)
    ar = [1, 2, 3, 4, 5]
    mad = mean_absolute_deviation(ar)
    sarr = array_to_string(ar)
    x = {"x": 1, "y": 2, "z": 3}
    y = {"m": 1, "n": 2, "o": 3}
    z = update_dict(x, y)



"""
Downloads files from Figshare.

Main page: https://figshare.com/authors/Kamal_Choudhary/4445539
"""

import zipfile
import tempfile
import os
import numpy as np
import io
import json
import requests
from jarvis.db.jsonutils import loadjson
from jarvis.io.vasp.outputs import Vasprun
from jarvis.io.vasp.inputs import Poscar
from jarvis.io.wannier.outputs import WannierHam
from tqdm import tqdm
import matplotlib.image as mpimg
from jarvis.analysis.stm.tersoff_hamann import TersoffHamannSTM

# import matplotlib.pyplot as plt
# plt.switch_backend("agg")


def get_db_info():
    """Get DB info."""
    db_info = {
        # https://doi.org/10.6084/m9.figshare.6815705
        "dft_2d": [
            "https://ndownloader.figshare.com/files/38521268",
            "d2-12-12-2022.json",
            "Obtaining 2D dataset 1.1k ...",
            "https://www.nature.com/articles/s41524-020-00440-1"
            + "\nOther versions:https://doi.org/10.6084/m9.figshare.6815705",
        ],
        # https://doi.org/10.6084/m9.figshare.6815699
        "dft_3d": [
            "https://ndownloader.figshare.com/files/38521619",
            "jdft_3d-12-12-2022.json",
            "Obtaining 3D dataset 76k ...",
            "https://www.nature.com/articles/s41524-020-00440-1"
            + "\nOther versions:https://doi.org/10.6084/m9.figshare.6815699",
        ],
        # https://doi.org/10.6084/m9.figshare.6815705
        "dft_2d_2021": [
            "https://ndownloader.figshare.com/files/26808917",
            "d2-3-12-2021.json",
            "Obtaining 2D dataset 1.1k ...",
            "https://www.nature.com/articles/s41524-020-00440-1",
        ],
        # https://doi.org/10.6084/m9.figshare.6815699
        "dft_3d_2021": [
            "https://ndownloader.figshare.com/files/29204826",
            "jdft_3d-8-18-2021.json",
            "Obtaining 3D dataset 55k ...",
            "https://www.nature.com/articles/s41524-020-00440-1",
        ],
        # https://doi.org/10.6084/m9.figshare.6815699
        "cfid_3d": [
            "https://ndownloader.figshare.com/files/29205201",
            "cfid_3d-8-18-2021.json",
            "Obtaining 3D dataset 55k ...",
            "https://www.nature.com/articles/s41524-020-00440-1"
            + "\nOther versions:https://doi.org/10.6084/m9.figshare.6815699",
        ],
        # https://doi.org/10.6084/m9.figshare.14213522
        "jff": [
            "https://ndownloader.figshare.com/files/28937793",
            # "https://ndownloader.figshare.com/files/26809760",
            "jff-7-24-2021.json",
            # "jff-3-12-2021.json",
            "Obtaining JARVIS-FF 2k ...",
            "https://www.nature.com/articles/s41524-020-00440-1",
        ],
        # https://doi.org/10.6084/m9.figshare.21667874
        "alignn_ff_db": [
            "https://ndownloader.figshare.com/files/38522315",
            # "https://ndownloader.figshare.com/files/26809760",
            "id_prop.json",
            "Obtaining ALIGNN-FF training DB 300k ...",
            "https://arxiv.org/abs/2209.05554",
        ],
        "mp_3d_2020": [
            "https://ndownloader.figshare.com/files/26791259",
            "all_mp.json",
            "Obtaining Materials Project-3D CFID dataset 127k...",
            "https://doi.org/10.1063/1.4812323",
        ],
        # https://doi.org/10.6084/m9.figshare.14177630
        "megnet": [
            "https://ndownloader.figshare.com/files/26724977",
            "megnet.json",
            "Obtaining MEGNET-3D CFID dataset 69k...",
            "https://pubs.acs.org/doi/10.1021/acs.chemmater.9b01294",
        ],
        # https://doi.org/10.6084/m9.figshare.14745435
        "megnet2": [
            "https://ndownloader.figshare.com/files/28332741",
            "megnet-mp-2019-04-01.json",
            "Obtaining MEGNET-3D CFID dataset 133k...",
            "https://pubs.acs.org/doi/10.1021/acs.chemmater.9b01294",
        ],
        # https://doi.org/10.6084/m9.figshare.14745327
        "edos_pdos": [
            "https://ndownloader.figshare.com/files/29216859",
            "edos-up_pdos-elast_interp-8-18-2021.json",
            "Interpolated electronic total dos spin-up dataset 55k...",
            "https://www.nature.com/articles/s41524-020-00440-1",
        ],
        # https://doi.org/10.6084/m9.figshare.13054247
        "mp_3d": [
            "https://ndownloader.figshare.com/files/24979850",
            "CFID_mp_desc_data_84k.json",
            "Obtaining Materials Project-3D CFID dataset 84k...",
            "https://doi.org/10.1063/1.4812323",
        ],
        # https://doi.org/10.6084/m9.figshare.13055333
        "oqmd_3d": [
            "https://ndownloader.figshare.com/files/24981170",
            "CFID_OQMD_460k.json",
            "Obtaining OQMD-3D CFID dataset 460k...",
            "https://www.nature.com/articles/npjcompumats201510",
        ],
        # https://doi.org/10.6084/m9.figshare.14206169
        "oqmd_3d_no_cfid": [
            "https://ndownloader.figshare.com/files/26790182",
            "all_oqmd.json",
            "Obtaining OQMD-3D  dataset 800k...",
            "https://www.nature.com/articles/npjcompumats201510",
        ],
        # https://doi.org/10.6084/m9.figshare.14205083
        "twod_matpd": [
            "https://ndownloader.figshare.com/files/26789006",
            "twodmatpd.json",
            "Obtaining 2DMatPedia dataset 6k...",
            "https://www.nature.com/articles/s41597-019-0097-3",
        ],
        # https://doi.org/10.6084/m9.figshare.14213603
        "polymer_genome": [
            "https://ndownloader.figshare.com/files/26809907",
            "pgnome.json",
            "Obtaining Polymer genome 1k...",
            "https://www.nature.com/articles/sdata201612",
        ],
        "qm9_std_jctc": [
            "https://ndownloader.figshare.com/files/28715319",
            "qm9_std_jctc.json",
            "Obtaining QM9 standardized dataset 130k,"
            + "From https://doi.org/10.1021/acs.jctc.7b00577,+",
            "https://www.nature.com/articles/sdata201422",
        ],
        # https://doi.org/10.6084/m9.figshare.14827584
        # Use qm9_std_jctc instaed
        "qm9_dgl": [
            "https://ndownloader.figshare.com/files/28541196",
            "qm9_dgl.json",
            "Obtaining QM9 dataset 130k, from DGL...",
            "https://www.nature.com/articles/sdata201422",
        ],
        # https://doi.org/10.6084/m9.figshare.14912820.v1
        "cod": [
            "https://ndownloader.figshare.com/files/28715301",
            "cod_db.json",
            "Obtaining COD dataset 431k",
            "https://doi.org/10.1107/S1600576720016532",
        ],
        # Use qm9_std_jctc instaed
        "qm9": [
            "https://ndownloader.figshare.com/files/27627596",
            "qm9_data_cfid.json",
            "Obtaining QM9 dataset 134k...",
            "https://www.nature.com/articles/sdata201422",
        ],
        # https://doi.org/10.6084/m9.figshare.15127788
        "qe_tb": [
            "https://ndownloader.figshare.com/files/29070555",
            "jqe_tb_folder.json",
            "Obtaining QETB dataset 860k...",
            "https://arxiv.org/abs/2112.11585",
        ],
        # https://doi.org/10.6084/m9.figshare.14812050
        "omdb": [
            "https://ndownloader.figshare.com/files/28501761",
            "omdbv1.json",
            "Obtaining OMDB dataset 12.5k...",
            "https://doi.org/10.1002/qute.201900023",
        ],
        # https://doi.org/10.6084/m9.figshare.14812044
        "qmof": [
            "https://figshare.com/ndownloader/files/30972640",
            "qmof_db.json",
            "Obtaining QMOF dataset 20k...",
            "https://www.cell.com/matter/fulltext/S2590-2385(21)00070-9",
        ],
        # https://doi.org/10.6084/m9.figshare.15127758
        "hmof": [
            "https://figshare.com/ndownloader/files/30972655",
            "hmof_db_9_18_2021.json",
            "Obtaining hMOF dataset 137k...",
            "https://doi.org/10.1021/acs.jpcc.6b08729",
        ],
        # https://figshare.com/account/projects/100325/articles/14960157
        "c2db": [
            "https://ndownloader.figshare.com/files/28682010",
            "c2db_atoms.json",
            "Obtaining C2DB dataset 3.5k...",
            "https://iopscience.iop.org/article/10.1088/2053-1583/aacfc1",
        ],
        # https://doi.org/10.6084/m9.figshare.25256236
        "halide_peroskites": [
            "https://figshare.com/ndownloader/files/44619562",
            "halide_peroskites.json",
            "Obtaining halide perovskite dataset229...",
            "https://doi.org/10.1039/D1EE02971A",
        ],
        # https://figshare.com/account/projects/100325/articles/14962356
        "hopv": [
            "https://ndownloader.figshare.com/files/28814184",
            "hopv_15.json",
            "Obtaining HOPV15 dataset 4.5k...",
            "https://www.nature.com/articles/sdata201686",
        ],
        # https://figshare.com/account/projects/100325/articles/14962356
        "pdbbind_core": [
            "https://ndownloader.figshare.com/files/28874802",
            "pdbbind_2015_core.json",
            "Obtaining PDBBind dataset 195...",
            "https://doi.org/10.1093/bioinformatics/btu626",
        ],
        # https://doi.org/10.6084/m9.figshare.14812038
        "pdbbind": [
            "https://ndownloader.figshare.com/files/28816368",
            "pdbbind_2015.json",
            "Obtaining PDBBind dataset 11k...",
            "https://doi.org/10.1093/bioinformatics/btu626",
        ],
        # https://doi.org/10.6084/m9.figshare.21713885
        "snumat": [
            "https://ndownloader.figshare.com/files/38521736",
            "snumat.json",
            "Obtaining SNUMAT Hybrid functional dataset 10k...",
            "https://www.nature.com/articles/s41597-020-00723-8",
        ],
        # https://doi.org/10.6084/m9.figshare.13215308
        "aflow2": [
            "https://ndownloader.figshare.com/files/25453265",
            "CFID_AFLOW2.json",
            "Obtaining AFLOW-2 CFID dataset 400k...",
            "https://doi.org/10.1016/j.commatsci.2012.02.005",
        ],
        # https://doi.org/10.6084/m9.figshare.14211860
        "arXiv": [
            "https://ndownloader.figshare.com/files/26804795",
            "arXivdataset.json",
            "Obtaining arXiv dataset 1.8 million...",
            "https://www.kaggle.com/Cornell-University/arxiv",
        ],
        # https://doi.org/10.6084/m9.figshare.14211857
        "cord19": [
            "https://ndownloader.figshare.com/files/26804798",
            "cord19.json",
            "Obtaining CORD19 dataset 223k...",
            "https://github.com/usnistgov/cord19-cdcs-nist",
        ],
        # https://doi.org/10.6084/m9.figshare.22583677
        "ssub": [
            "https://figshare.com/ndownloader/files/40084921",
            "ssub.json",
            "Obtaining SSUB dataset 1726...",
            "https://github.com/wolverton-research-group/qmpy",
        ],
        # https://doi.org/10.6084/m9.figshare.22721047
        "mlearn": [
            # "https://figshare.com/ndownloader/files/40424156",
            "https://figshare.com/ndownloader/files/40357663",
            "mlearn.json",
            "Obtaining mlearn dataset 1730...",
            "https://github.com/materialsvirtuallab/mlearn",
        ],
        # https://doi.org/10.6084/m9.figshare.22814318
        "foundry_ml_exp_bandgaps": [
            "https://figshare.com/ndownloader/files/40557743",
            "foundry_ml_exp_bandgaps.json",
            "Obtaining foundry_ml_exp_bandgaps dataset 2069...",
            "https://foundry-ml.org/#/datasets/10.18126/wg3u-g8vu",
        ],
        # ToFix# https://doi.org/10.6084/m9.figshare.22815926
        # "mat_scholar_ner": [
        #    "https://figshare.com/ndownloader/files/40563593",
        #    "mat_scholar_ner.json",
        #    "Obtaining mat_scholar_ner dataset XYZ...",
        #    "https://pubs.acs.org/doi/10.1021/acs.jcim.9b00470",
        # ],
        # https://doi.org/10.6084/m9.figshare.22817633
        # Contains repeats
        "ocp10k": [
            "https://figshare.com/ndownloader/files/40566122",
            "ocp10k.json",
            "Obtaining OCP 10k train dataset, 59886...",
            "https://github.com/Open-Catalyst-Project/ocp",
        ],
        # https://doi.org/10.6084/m9.figshare.22817651
        "arxiv_summary": [
            "https://figshare.com/ndownloader/files/40566137",
            "arxiv_summary.json",
            "Obtaining arxiv summary cond.mat dataset 137927...",
            "https://github.com/usnistgov/chemnlp",
        ],
        # TODO:PubChem
        # https://doi.org/10.6084/m9.figshare.22975787
        "supercon_chem": [
            "https://figshare.com/ndownloader/files/40719260",
            "supercon_chem.json",
            "Obtaining supercon chem dataset 16414...",
            "https://www.nature.com/articles/s41524-018-0085-8",
        ],
        # https://doi.org/10.6084/m9.figshare.22976285
        "mag2d_chem": [
            "https://figshare.com/ndownloader/files/40720004",
            "mag2d_chem.json",
            "Obtaining magnetic 2D chem dataset 226...",
            "https://doi.org/10.24435/materialscloud:2019.0020/v1",
        ],
        # https://doi.org/10.6084/m9.figshare.23000573
        "vacancydb": [
            "https://figshare.com/ndownloader/files/40750811",
            "vacancydb.json",
            "Obtaining vacancy dataset 464...",
            "https://arxiv.org/abs/2205.08366",
        ],
        # Contains repeats
        # https://doi.org/10.6084/m9.figshare.23206193
        "ocp100k": [
            "https://figshare.com/ndownloader/files/40902845",
            "ocp100k.json",
            "Obtaining OCP100k dataset 149886...",
            "https://github.com/Open-Catalyst-Project/ocp",
        ],
        # https://doi.org/10.6084/m9.figshare.23250629
        "ocp_all": [
            "https://figshare.com/ndownloader/files/40974599",
            "ocp_all.json",
            "Obtaining OCPall dataset 510214...",
            "https://github.com/Open-Catalyst-Project/ocp",
        ],
        # https://doi.org/10.6084/m9.figshare.23225687
        "tinnet_N": [
            "https://figshare.com/ndownloader/files/40934285",
            "tinnet_N.json",
            "Obtaining TinNet Nitrogen dataset 329...",
            "https://github.com/hlxin/tinnet",
        ],
        # https://doi.org/10.6084/m9.figshare.23254151
        "tinnet_O": [
            "https://figshare.com/ndownloader/files/40978943",
            "tinnet_O.json",
            "Obtaining TinNet Oxygen dataset 747...",
            "https://github.com/hlxin/tinnet",
        ],
        # https://doi.org/10.6084/m9.figshare.23254154
        "tinnet_OH": [
            "https://figshare.com/ndownloader/files/40978949",
            "tinnet_OH.json",
            "Obtaining TinNet OH dataset 748...",
            "https://github.com/hlxin/tinnet",
        ],
        # https://doi.org/10.6084/m9.figshare.23909478
        "AGRA_O": [
            "https://figshare.com/ndownloader/files/41923284",
            "AGRA_O.json",
            "Obtaining AGRA Oxygen dataset 1000...",
            "https://github.com/Feugmo-Group/AGRA",
        ],
        # https://doi.org/10.6084/m9.figshare.23909478
        "AGRA_OH": [
            "https://figshare.com/ndownloader/files/41923287",
            "AGRA_OH.json",
            "Obtaining AGRA OH dataset 875...",
            "https://github.com/Feugmo-Group/AGRA",
        ],
        # https://doi.org/10.6084/m9.figshare.23909478
        "AGRA_CO": [
            "https://figshare.com/ndownloader/files/41923278",
            "AGRA_CO.json",
            "Obtaining AGRA CO dataset 193...",
            "https://github.com/Feugmo-Group/AGRA",
        ],
        # https://doi.org/10.6084/m9.figshare.23909478
        "AGRA_CHO": [
            "https://figshare.com/ndownloader/files/41923275",
            "AGRA_CHO.json",
            "Obtaining AGRA Oxygen dataset 214...",
            "https://github.com/Feugmo-Group/AGRA",
        ],
        # https://doi.org/10.6084/m9.figshare.23909478
        "AGRA_COOH": [
            "https://figshare.com/ndownloader/files/41923281",
            "AGRA_COOH.json",
            "Obtaining AGRA COOH dataset 280...",
            "https://github.com/Feugmo-Group/AGRA",
        ],
        # https://doi.org/10.6084/m9.figshare.21370572
        "supercon_3d": [
            "https://figshare.com/ndownloader/files/38307921",
            "jarvis_epc_data_figshare_1058.json",
            "Obtaining supercond. Tc dataset 1058...",
            "https://www.nature.com/articles/s41524-022-00933-1",
        ],
        # https://doi.org/10.6084/m9.figshare.21370572
        "supercon_2d": [
            "https://figshare.com/ndownloader/files/38950433",
            "jarvis_epc_data_2d.json",
            "Obtaining supercond. Tc dataset 161...",
            "https://doi.org/10.1021/acs.nanolett.2c04420",
        ],
        # https://doi.org/10.6084/m9.figshare.23267852
        "m3gnet_mpf": [
            "https://figshare.com/ndownloader/files/41009036",
            "m3gnet_mpf.json",
            "Obtaining m3gnet_mpf dataset 168917...",
            "https://github.com/materialsvirtuallab/m3gnet",
        ],
        # https://doi.org/10.6084/m9.figshare.13154159
        "raw_files": [
            "https://ndownloader.figshare.com/files/25295732",
            "figshare_data-10-28-2020.json",
            "Obtaining raw io files 145k...",
            "https://www.nature.com/articles/s41524-020-00440-1",
        ],
    }
    return db_info


# Format: download_link, filename, message, reference
# Figshare link: https://figshare.com/account/home#/projects/100325


def get_stm_2d_dataset():
    """Get 2D STM image dataset."""
    # Ref: https://www.nature.com/articles/s41597-021-00824-y
    link_1 = "https://ndownloader.figshare.com/files/21884952"
    r_jpg = requests.get(link_1)
    z = zipfile.ZipFile(io.BytesIO(r_jpg.content))
    link_2 = "https://ndownloader.figshare.com/files/21893379"
    r_json = requests.get(link_2).content
    latts = json.loads(r_json)
    namelist = z.namelist()
    pos_bias = []
    neg_bias = []

    print("Obtaining 2D STM dataset ...")
    for i in namelist:
        img_str = z.read(i)
        values = mpimg.imread(io.BytesIO((img_str)), format="jpg")
        # img=Image(values=values)
        jid = i.split("/")[-1].split("_")[0]
        bias = i.split("/")[-1].split("_")[1].split(".jpg")[0]
        lat_system = latts[jid]
        if bias == "pos":
            info = {}
            info["jid"] = jid
            info["image_values"] = values
            info["lat_type"] = lat_system
            pos_bias.append(info)
        if bias == "neg":
            info = {}
            info["jid"] = jid
            info["image_values"] = values
            info["lat_type"] = lat_system
            neg_bias.append(info)
    return pos_bias, neg_bias


def get_request_data(
    js_tag="jdft_2d-4-26-2020.json",
    url="https://ndownloader.figshare.com/files/22471019",
    store_dir=None,
):
    """Get data with progress bar."""
    zfile = js_tag + ".zip"
    if store_dir is None:
        path = str(os.path.join(os.path.dirname(__file__), zfile))
    else:
        path = str(os.path.join(store_dir, zfile))

    # path = str(os.path.join(os.path.dirname(__file__), js_tag))
    if not os.path.isfile(path):
        # zfile = str(os.path.join(os.path.dirname(__file__), "tmp.zip"))
        response = requests.get(url, stream=True)
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024  # 1 Kibibyte
        progress_bar = tqdm(
            total=total_size_in_bytes, unit="iB", unit_scale=True
        )
        with open(path, "wb") as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        # f = open(zfile, "wb")
        # f.write(r.content)
        # f.close()
    print("Loading the zipfile...")
    data = json.loads(zipfile.ZipFile(path).read(js_tag))
    print("Loading completed.")

    #    with zipfile.ZipFile(zfile, "r") as zipObj:
    #        # zipObj.extract(path)
    #        zipObj.extractall(os.path.join(os.path.dirname(__file__)))
    #    os.remove(zfile)
    # data = loadjson(path)
    return data


def data(dataset="dft_2d", store_dir=None):
    """Provide main function to download datasets."""
    db_info = get_db_info()
    if dataset not in list(db_info.keys()):
        raise ValueError("Check DB name options.")
    url = db_info[dataset][0]
    js_tag = db_info[dataset][1]
    message = db_info[dataset][2]
    print(message)
    message = "Reference:" + db_info[dataset][3]
    print(message)
    # r = requests.get(url)
    # z = zipfile.ZipFile(io.BytesIO(r.content))
    # data = json.loads(z.read(js_tag).decode("utf-8"))

    # r = requests.get(url)
    # z = zipfile.ZipFile(io.BytesIO(r.content))
    # wdat = z.read(js_tag).decode("utf-8")
    # fd, path = tempfile.mkstemp()
    # with os.fdopen(fd, "w") as tmp:
    #    tmp.write(wdat)
    # data = loadjson(path)

    # path = str(os.path.join(os.path.dirname(__file__), js_tag))
    # if not os.path.isfile(path):
    #    zfile = str(os.path.join(os.path.dirname(__file__), "tmp.zip"))
    #    r = requests.get(url)
    #    f = open(zfile, "wb")
    #    f.write(r.content)
    #    f.close()

    #    with zipfile.ZipFile(zfile, "r") as zipObj:
    #        # zipObj.extract(path)
    #        zipObj.extractall(os.path.join(os.path.dirname(__file__)))
    #    os.remove(zfile)
    # data = loadjson(path)
    dat = get_request_data(js_tag=js_tag, url=url, store_dir=store_dir)
    return dat


def get_jid_data(jid="JVASP-667", dataset="dft_2d"):
    """Get info for a jid and dataset."""
    d = data(dataset)
    for i in d:
        if i["jid"] == jid:
            return i


def get_ff_eneleast():
    """Get JARVIS-FF related data."""
    jff1 = str(os.path.join(os.path.dirname(__file__), "jff1.json"))
    if not os.path.isfile(jff1):
        r = requests.get("https://ndownloader.figshare.com/files/10307139")
        f = open(jff1, "wb")
        f.write(r.content)
        f.close()
    data_ff1 = loadjson(jff1)
    return data_ff1


def make_stm_from_prev_parchg(
    jid="JVASP-667", bias="Negative", filename="stm_image.png", min_size=10
):
    """Make STM images from previously calculated PARVHG files for 2D."""
    fls = data("raw_files")
    for i in fls["STM"]:
        zip_name = jid + "_" + bias + ".zip"
        if i["name"] == zip_name:
            zip_file_url = i["download_url"]
            r = requests.get(zip_file_url)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            pchg = z.read("PARCHG").decode("utf-8")
            fd, path = tempfile.mkstemp()
            with os.fdopen(fd, "w") as tmp:
                tmp.write(pchg)
            TH_STM = TersoffHamannSTM(
                chg_name=path, min_size=min_size, zcut=None
            )
            t_height = TH_STM.constant_height(filename=filename)
            print("t_height", t_height)
            return i


def get_wann_electron(jid="JVASP-816"):
    """Download electron WTBH if available."""
    w = ""
    ef = ""
    fls = data("raw_files")
    for i in fls["WANN"]:
        if i["name"].split(".zip")[0] == jid:
            r = requests.get(i["download_url"])
            z = zipfile.ZipFile(io.BytesIO(r.content))
            wdat = z.read("wannier90_hr.dat").decode("utf-8")
            js_file = jid + ".json"
            js = z.read(js_file).decode("utf-8")
            fd, path = tempfile.mkstemp()
            with os.fdopen(fd, "w") as tmp:
                tmp.write(wdat)
            w = WannierHam(path)
            fd, path = tempfile.mkstemp()
            with os.fdopen(fd, "w") as tmp:
                tmp.write(js)
            d = loadjson(path)
            ef = d["info_mesh"]["efermi"]
            fd, path = tempfile.mkstemp()
            pos = z.read("POSCAR").decode("utf-8")
            with os.fdopen(fd, "w") as tmp:
                tmp.write(pos)
            atoms = Poscar.from_file(path).atoms
    return w, ef, atoms


def get_wann_phonon(jid="JVASP-1002", factor=15.633302):
    """Download phonon WTBH if available."""
    # Requires phonopy
    from jarvis.io.phonopy.outputs import get_phonon_tb

    fls = data("raw_files")

    for i in fls["FD-ELAST"]:
        if isinstance(i, dict):
            if i["name"].split(".zip")[0] == jid:
                r = requests.get(i["download_url"])
                z = zipfile.ZipFile(io.BytesIO(r.content))
                vrun_path = z.read("vasprun.xml").decode("utf-8")
                fd, path = tempfile.mkstemp()
                with os.fdopen(fd, "w") as tmp:
                    tmp.write(vrun_path)
                vrun = Vasprun(path)
                fc = vrun.phonon_data()["force_constants"]
                atoms = vrun.all_structures[0]
                # print(atoms)
                # atoms = Atoms.from_poscar(pos)
                # print(atoms)
                fd, path = tempfile.mkstemp()
                get_phonon_tb(fc=fc, atoms=atoms, out_file=path, factor=factor)
                # cvn = Spacegroup3D(atoms).conventional_standard_structure
                w = WannierHam(path)
                return w, atoms


def get_hk_tb(k=np.array([0, 0, 0]), w=[]):
    """Get Wannier TB Hamiltonian."""
    nr = w.R.shape[0]
    hk = np.zeros((w.nwan, w.nwan), dtype=complex)
    kmat = np.tile(k, (nr, 1))
    exp_ikr = np.exp(1.0j * 2 * np.pi * np.sum(kmat * w.R, 1))
    temp = np.zeros(w.nwan**2, dtype=complex)
    for i in range(nr):
        temp += exp_ikr[i] * w.HR[i, :]
    hk = np.reshape(temp, (w.nwan, w.nwan))
    hk = (hk + hk.T.conj()) / 2.0
    return hk


"""
QM9 xyz file
>>> def get_val(filename=''):
        f=open(filename,'r')
        lines=f.read().splitlines()
        f.close()
        info={}
        for i in range(len(attr_index)):
           info[attr_index[i]]=float(lines[1].split()[i+2])
        return info
"""

"""
if __name__ == "__main__":

    data_2d = data(dataset='dft_2d')
    print('2d',len(data_2d))
    data_3d = data(dataset='dft_3d')
    print('3d',len(data_3d))
    data_ml = data(dataset='cfid_3d')
    print('cfid3d',len(data_ml))
    data_ff = get_ff_eneleast()
    print ('ff',len(data_ff))
"""
